import numpy as np
import os
import shutil

from json import dump, load
from subprocess import call
from typing import List, Union

from pymatgen.io.vasp.inputs import Poscar
from pymatgen.io.vasp.outputs import Vasprun
from pymatgen.io.vasp.sets import VaspInputSet
from pymatgen.core.periodic_table import Element

from flare.struc import Structure
from flare.utils.element_coder import NumpyEncoder

name = "VASP"


def check_vasprun(vasprun: Union[str, Vasprun], vasprun_kwargs: dict = {}) -> Vasprun:
    """
    Helper utility to take a vasprun file name or Vasprun object
    and return a vasprun object.
    :param vasprun: vasprun filename or object
    """

    if type(vasprun) == str:
        return Vasprun(vasprun, **vasprun_kwargs)
    elif type(vasprun) == Vasprun:
        return vasprun
    else:
        raise ValueError("Vasprun argument is not a string or Vasprun instance!")


def parse_dft_input(input: str):
    """
    Returns the positions, species, and cell of a POSCAR file.
    Outputs are specced for OTF module.

    :param input: POSCAR file input
    :return:
    """

    pmg_structure = Poscar.from_file(input).structure
    flare_structure = Structure.from_pmg_structure(pmg_structure)

    positions = flare_structure.positions
    species = flare_structure.species_labels
    cell = flare_structure.cell
    # TODO Allow for custom masses in POSCAR

    elements = set(species)

    # conversion from amu to md units
    mass_dict = {elt: Element(elt).atomic_mass * 0.000103642695727 for elt in elements}

    return positions, species, cell, mass_dict


def run_dft(
    calc_dir: str,
    dft_loc: str,
    structure: Structure = None,
    en: bool = False,
    vasp_cmd="{}",
):
    """
    Run a VASP calculation.
    :param calc_dir: Name of directory to perform calculation in
    :param dft_loc: Name of VASP command (i.e. executable location)
    :param structure: flare structure object
    :param en: whether to return the final energy along with the forces
    :param vasp_cmd: Command to run VASP (leave a formatter "{}" to insert dft_loc);
        this can be used to specify mpirun, srun, etc.
    Returns:
        forces on each atom (and energy if en=True)
    """

    currdir = os.getcwd()
    if structure:
        edit_dft_input_positions("POSCAR", structure)
    try:
        if currdir != calc_dir:
            os.chdir(calc_dir)
        call(vasp_cmd.format(dft_loc).split())

        if en:
            parse_func = parse_dft_forces_and_energy
        else:
            parse_func = parse_dft_forces

        try:
            forces = parse_func("vasprun.xml")
        except FileNotFoundError:
            os.chdir(currdir)
            raise FileNotFoundError(
                "Could not load vasprun.xml."
                "The calculation may not have finished."
                f"Current directory is {os.getcwd()}"
            )

        os.chdir(currdir)
        return forces

    except Exception as e:
        os.chdir(currdir)
        raise e


def run_dft_par(
    dft_input: str,
    structure: Structure,
    dft_command: str = None,
    n_cpus=1,
    dft_out="vasprun.xml",
    parallel_prefix="mpi",
    mpi=None,
    npool=None,
    screen_out="vasp.out",
    **dft_kwargs,
):
    # TODO Incorporate Custodian.
    edit_dft_input_positions(dft_input, structure)

    if dft_command is None or not os.environ.get("VASP_COMMAND"):
        raise FileNotFoundError(
            "Warning: No VASP Command passed, or stored in "
            "environment as VASP_COMMAND. "
        )

    if n_cpus > 1:
        # why is parallel prefix needed?
        if (parallel_prefix == "mpi") and (mpi == "mpi"):
            dft_command = f"mpirun -np {n_cpus} {dft_command}"
        elif mpi == "srun":
            dft_command = f"srun -n {n_cpus} {dft_command}"

    else:
        serial_prefix = dft_kwargs.get("serial_prefix", "")
        dft_command = f"{serial_prefix} {dft_command}"

    with open(screen_out, "w+") as fout:
        call(dft_command.split(), stdout=fout)

    return parse_dft_forces(dft_out)


def dft_input_to_structure(poscar: str):
    """
    Parse the DFT input in a directory.
    :param vasp_input: directory of vasp input
    """
    return Structure.from_pmg_structure(Poscar.from_file(poscar).structure)


def edit_dft_input_positions(output_name: str, structure: Structure):
    """
    Writes a VASP POSCAR file from structure with the name poscar .
    WARNING: Destructively replaces the file with the name specified by poscar
    :param poscar: Name of file
    :param structure: structure to write to file
    """
    if os.path.isfile(output_name):
        shutil.copyfile(output_name, output_name + ".bak")
    poscar = Poscar(structure.to_pmg_structure())
    poscar.write_file(output_name)
    return output_name


def parse_dft_forces(vasprun: Union[str, Vasprun]):
    """
    Parses the DFT forces from the last ionic step of a VASP vasprun.xml file
    :param vasprun: pymatgen Vasprun object or vasprun filename
    """
    vasprun = check_vasprun(vasprun)
    istep = vasprun.ionic_steps[-1]
    return np.array(istep["forces"])


def parse_dft_forces_and_energy(vasprun: Union[str, Vasprun]):
    """
    Parses the DFT forces and energy from a VASP vasprun.xml file
    :param vasprun: pymatgen Vasprun object or vasprun filename
    """
    vasprun = check_vasprun(vasprun)
    istep = vasprun.ionic_steps[-1]
    return np.array(istep["forces"]), istep["electronic_steps"][-1]["e_0_energy"]


def md_trajectory_from_vasprun(
    vasprun: Union[str, Vasprun], ionic_step_skips=1, vasprun_kwargs: dict = {}
):
    """
    Returns a list of flare Structure objects decorated with forces, stress,
    and total energy from a MD trajectory performed in VASP.
    :param vasprun: pymatgen Vasprun object or vasprun filename
    :param ionic_step_skips: if True, only samples the configuration every
        ionic_skip_steps steps.
    """
    vasprun = check_vasprun(vasprun, vasprun_kwargs)

    struc_lst = []
    for step in vasprun.ionic_steps[::ionic_step_skips]:
        structure = Structure.from_pmg_structure(step["structure"])
        structure.energy = step["electronic_steps"][-1]["e_0_energy"]
        # TODO should choose e_wo_entrp or e_fr_energy?
        structure.forces = np.array(step.get("forces"))
        structure.stress = np.array(step.get("stress"))
        struc_lst.append(structure)

    return struc_lst



"""The :class:`AtomicEnvironment` object stores information about the local
environment of an atom. :class:`AtomicEnvironment` objects are inputs to the
2-, 3-, and 2+3-body kernels."""

from typing import List, Union
from copy import deepcopy
from math import ceil
from json import dumps, loads

import pickle as pickle
import numpy as np

import flare.kernels.cutoffs as cf
from flare.utils import NumpyEncoder
from flare.struc import Structure
from flare.utils.env_getarray import (
    get_2_body_arrays,
    get_3_body_arrays,
    get_m2_body_arrays,
)


class AtomicEnvironment:
    """Contains information about the local environment of an atom,
    including arrays of pair and triplet distances and the chemical
    species of atoms in the environment.

    :param structure: Structure of atoms.
    :type structure: struc.Structure
    :param atom: Index of the atom in the structure.
    :type atom: int
    :param cutoffs: 2- and 3-body cutoff radii. 2-body if one cutoff is
    given, 2+3-body if two are passed.
    :type cutoffs: np.ndarray
    :param cutoffs_mask: a dictionary to store multiple cutoffs if neede
                         it should be exactly the same as the hyps mask
    :type cutoffs_mask: dict

    The cutoffs_mask allows the user to define multiple cutoffs for different
    bonds, triples, and many body interaction. This dictionary should be
    consistent with the hyps_mask used in the GuassianProcess object.

    * specie_mask: 118-long integer array descirbing which elements belong to
        like groups for determining which bond hyperparameters to use.
        For instance, [0,0,1,1,0 ...] assigns H to group 0, He and
        Li to group 1, and Be to group 0 (the 0th register is ignored).
    * nspecie: Integer, number of different species groups (equal to number of
        unique values in specie_mask).
    * ntwobody: Integer, number of different hyperparameter/cutoff sets to
        associate with different 2-body pairings of atoms in groups defined in
        specie_mask.
    * twobody_mask: Array of length nspecie^2, which describes the cutoff to
        associate with different pairings of species types. For example, if
        there are atoms of type 0 and 1, then twobody_mask defines which cutoff
        to use for parings [0-0, 0-1, 1-0, 1-1]: if we wanted cutoff0 for
        0-0 parings and set 1 for 0-1 and 1-1 pairings, then we would make
        twobody_mask [0, 1, 1, 1].
    * twobody_cutoff_list: Array of length ntwobody, which stores the cutoff
        used for different types of bonds defined in twobody_mask
    * ncut3b: Integer, number of different cutoffs sets to associate
        with different 3-body pariings of atoms in groups defined in
        specie_mask.
    * cut3b_mask: Array of length nspecie^2, which describes the cutoff to
        associate with different bond types in triplets. For example, in a
        triplet (C, O, H) , there are three cutoffs. Cutoffs for CH bond, CO
        bond and OH bond. If C and O are associate with atom group 1 in
        specie_mask and H are associate with group 0 in specie_mask, the
        cut3b_mask[1*nspecie+0] determines the C/O-H bond cutoff, and
        cut3b_mask[1*nspecie+1] determines the C-O bond cutoff. If we want the
        former one to use the 1st cutoff in threebody_cutoff_list and the later
        to use the 2nd cutoff in threebody_cutoff_list, the cut3b_mask should
        be [0, 0, 0, 1].
    * threebody_cutoff_list: Array of length ncut3b, which stores the cutoff
        used for different types of bonds in triplets.
    * nmanybody: Integer, number of different cutoffs set to associate with
        different coordination numbers.
    * manybody_mask: Similar to twobody_mask and cut3b_mask.
    * manybody_cutoff_list: Array of length nmanybody, stores the cutoff used
        for different many body terms

    Examples can be found at the end of in tests/test_env.py

    """

    all_kernel_types = ["twobody", "threebody", "manybody"]
    ndim = {"twobody": 2, "threebody": 3, "manybody": 2, "cut3b": 2}

    def __init__(self, structure: Structure, atom: int, cutoffs, cutoffs_mask=None):

        self.structure = structure
        self.positions = structure.wrapped_positions
        self.cell = np.array(structure.cell)
        self.species = structure.coded_species

        # backward compatability
        if not isinstance(cutoffs, dict):
            newcutoffs = {"twobody": cutoffs[0]}
            if len(cutoffs) > 1:
                newcutoffs["threebody"] = cutoffs[1]
            if len(cutoffs) > 2:
                newcutoffs["manybody"] = cutoffs[2]
            cutoffs = newcutoffs

        if cutoffs_mask is None:
            cutoffs_mask = {"cutoffs": cutoffs}
        elif cutoffs is not None:
            cutoffs_mask["cutoffs"] = deepcopy(cutoffs)

        # Set the sweep array based on the max cutoff.
        sweep_val = ceil(np.max(list(cutoffs.values())) / structure.max_cutoff)
        self.sweep_val = sweep_val
        self.sweep_array = np.arange(-sweep_val, sweep_val + 1, 1)

        self.atom = atom
        self.ctype = structure.coded_species[atom]

        self.twobody_cutoff = 0
        self.threebody_cutoff = 0
        self.manybody_cutoff = 0

        self.ntwobody = 1
        self.ncut3b = 0
        self.nmanybody = 0

        self.nspecie = 1
        self.specie_mask = None
        self.twobody_mask = None
        self.threebody_mask = None
        self.manybody_mask = None
        self.twobody_cutoff_list = None
        self.threebody_cutoff_list = None
        self.manybody_cutoff_list = None

        self.setup_mask(cutoffs_mask)

        assert (
            self.threebody_cutoff <= self.twobody_cutoff
        ), "2b cutoff has to be larger than 3b cutoff"
        # # TO DO, once the mb function is updated to use the bond_array_2
        # # this block should be activated.
        # assert self.manybody_cutoff <= self.twobody_cutoff, \
        #         "mb cutoff has to be larger than mb cutoff"

        self.bond_array_2 = None
        self.etypes = None
        self.bond_inds = None
        self.bond_array_3 = None
        self.cross_bond_inds = None
        self.cross_bond_dists = None
        self.triplet_counts = None
        self.q_array = None
        self.q_neigh_array = None
        self.q_grads = None
        self.q_neigh_grads = None
        self.unique_species = None
        self.etypes_mb = None

        self.compute_env()

    def setup_mask(self, cutoffs_mask):

        self.cutoffs_mask = deepcopy(cutoffs_mask)
        self.cutoffs = cutoffs_mask["cutoffs"]

        for kernel in AtomicEnvironment.all_kernel_types:
            ndim = AtomicEnvironment.ndim[kernel]
            if kernel in self.cutoffs:
                setattr(self, kernel + "_cutoff", self.cutoffs[kernel])

        if self.twobody_cutoff == 0:
            self.twobody_cutoff = np.max([self.threebody_cutoff, self.manybody_cutoff])

            self.cutoffs["twobody"] = self.twobody_cutoff

        self.nspecie = cutoffs_mask.get("nspecie", 1)
        if "specie_mask" in cutoffs_mask:
            self.specie_mask = np.array(cutoffs_mask["specie_mask"], dtype=np.int)

        for kernel in AtomicEnvironment.all_kernel_types:
            ndim = AtomicEnvironment.ndim[kernel]
            if kernel in self.cutoffs:
                setattr(self, kernel + "_cutoff", self.cutoffs[kernel])
                setattr(self, "n" + kernel, 1)
                if kernel != "threebody":
                    name_list = [
                        kernel + "_cutoff_list",
                        "n" + kernel,
                        kernel + "_mask",
                    ]
                    for name in name_list:
                        if name in cutoffs_mask:
                            setattr(self, name, cutoffs_mask[name])
                else:
                    self.ncut3b = cutoffs_mask.get("ncut3b", 1)
                    self.cut3b_mask = cutoffs_mask.get("cut3b_mask", None)
                    if "threebody_cutoff_list" in cutoffs_mask:
                        self.threebody_cutoff_list = np.array(
                            cutoffs_mask["threebody_cutoff_list"], dtype=np.float
                        )

    def compute_env(self):

        # get 2-body arrays
        if self.ntwobody >= 1:
            bond_array_2, bond_positions_2, etypes, bond_inds = get_2_body_arrays(
                self.positions,
                self.atom,
                self.cell,
                self.twobody_cutoff,
                self.twobody_cutoff_list,
                self.species,
                self.sweep_array,
                self.nspecie,
                self.specie_mask,
                self.twobody_mask,
            )

            self.bond_array_2 = bond_array_2
            self.etypes = etypes
            self.bond_inds = bond_inds

        # if 2 cutoffs are given, create 3-body arrays
        if self.ncut3b > 0:
            (
                bond_array_3,
                cross_bond_inds,
                cross_bond_dists,
                triplet_counts,
            ) = get_3_body_arrays(
                bond_array_2,
                bond_positions_2,
                self.species[self.atom],
                etypes,
                self.threebody_cutoff,
                self.threebody_cutoff_list,
                self.nspecie,
                self.specie_mask,
                self.cut3b_mask,
            )

            self.bond_array_3 = bond_array_3
            self.cross_bond_inds = cross_bond_inds
            self.cross_bond_dists = cross_bond_dists
            self.triplet_counts = triplet_counts

        # if 3 cutoffs are given, create many-body arrays
        if self.nmanybody > 0:
            (
                self.q_array,
                self.q_neigh_array,
                self.q_grads,
                self.q_neigh_grads,
                self.unique_species,
                self.etypes_mb,
            ) = get_m2_body_arrays(
                self.positions,
                self.atom,
                self.cell,
                self.manybody_cutoff,
                self.manybody_cutoff_list,
                self.species,
                self.sweep_array,
                self.nspecie,
                self.specie_mask,
                self.manybody_mask,
                cf.quadratic_cutoff,
            )

    def as_str(self) -> str:
        """
        Returns string dictionary serialization cast as string.

        :return: output of as_dict method cast as string
        :rtype: str
        """
        return dumps(self.as_dict(), cls=NumpyEncoder)

    def as_dict(self, include_structure: bool = False):
        """
        Returns Atomic Environment object as a dictionary for serialization
        purposes. Optional to not include the structure to avoid redundant
        information.
        :return:
        """

        dictionary = dict(vars(self))
        dictionary["object"] = "AtomicEnvironment"
        dictionary["forces"] = self.structure.forces

        # Backward compatibility for older models: Cutoffs mask.
        # Can be deleted one day if support is dropped for older (Pre June
        # 2020) pickled environment objects.
        cutoffs_mask = getattr(self, "cutoffs_mask", {"cutoffs": self.cutoffs})
        if not hasattr(self, "cutoffs_mask"):
            self.cutoffs_mask = cutoffs_mask
        dictionary["cutoffs_mask"] = cutoffs_mask

        if not include_structure:
            del dictionary["structure"]
        else:
            dictionary["structure"] = dictionary["structure"].as_dict()

        return dictionary

    @staticmethod
    def from_dict(dictionary):
        """
        Loads in atomic environment object from a dictionary which was
        serialized by the to_dict method.

        :param dictionary: Dictionary describing atomic environment.
        """
        # TODO Instead of re-computing 2 and 3 body environment,
        # directly load in, this would be much more efficient

        struc = Structure(
            cell=np.array(dictionary["cell"]),
            positions=dictionary["positions"],
            species=dictionary["species"],
        )
        index = dictionary["atom"]

        if dictionary.get("cutoffs") is not None:
            cutoffs = dictionary["cutoffs"]
        else:
            cutoffs = {}

        cutoffs_mask = dictionary.get("cutoffs_mask", None)

        return AtomicEnvironment(struc, index, cutoffs, cutoffs_mask=cutoffs_mask)

    @staticmethod
    def from_file(
        file_name: str, format: str = ""
    ) -> Union["flare.env.AtomicEnvironment", List["flare.env.AtomicEnvironment"]]:
        """
        Load an atomic environment from a file or a series of atomic environments
        :param file_name:
        :param format:
        :return:
        """

        # Ensure the file specified exists.
        with open(file_name, "r") as _:
            pass

        if "pickle" in file_name or (
            "pickle" in format.lower() or "binary" in format.lower()
        ):
            with open(file_name, "rb") as f:
                return pickle.load(f)

        if "json" in format.lower() or ".json" in file_name:
            # Assumed format is one atomic environment per line,
            # or one line with many atomic environments
            with open(file_name, "r") as f:
                thelines = f.readlines()

                non_empty_lines = [loads(line) for line in thelines if len(line) > 2]

            envs = [
                AtomicEnvironment.from_dict(env_dict) for env_dict in non_empty_lines
            ]

            if len(envs) == 1:
                return envs[0]
            else:
                return envs

    def __str__(self):
        atom_type = self.ctype
        neighbor_types = self.etypes
        n_neighbors = len(self.bond_array_2)
        string = "Atomic Env. of Type {} surrounded by {} atoms of Types {}".format(
            atom_type, n_neighbors, sorted(list(set(neighbor_types)))
        )

        return string

    @property
    def force(self) -> "np.ndarray":
        return self.structure.forces[self.atom]



"""
Tool to enable the development of a GP model based on an AIMD
trajectory with many customizable options for fine control of training.
Contains methods to transfer the model to an OTF run or MD engine run.


Seed frames
-----------
The various parameters in the :class:`TrajectoryTrainer` class related to
"Seed frames" are to help you  train a model which does not yet have a
training set. Uncertainty- and force-error driven training will go better with
a somewhat populated training set, as force and uncertainty estimates
are better behaveed with more data.

You may pass in a set of seed frames or atomic environments.
All seed environments will be added to the GP model; seed frames will
be iterated through and atoms will be added at random.
There are a few reasons why you would want to pay special attention to an
individual species.

If you are studying a system where the dynamics of one species are
particularly important and so you want a good representation in the training
set, then you would want to include as many as possible in the training set
during the seed part of the training.

Inversely, if a system has high representation of a species well-described
by a simple 2+3 body kernel, you may want it to be less well represented
in the seeded training set.

By specifying the pre_train_atoms_per_element, you can limit the number of
atoms of a given species which are added in. You can also limit the number
of atoms which are added from a given seed frame.

"""
import json as json
import logging
import numpy as np
import time
import warnings

from copy import deepcopy
from math import inf
from typing import List, Tuple, Union

from flare.env import AtomicEnvironment
from flare.gp import GaussianProcess
from flare.mgp.mgp import MappedGaussianProcess
from flare.output import Output
from flare.predict import (
    predict_on_atom,
    predict_on_atom_en,
    predict_on_structure_par,
    predict_on_structure_par_en,
    predict_on_structure_mgp,
)
from flare.struc import Structure
from flare.utils.element_coder import element_to_Z, Z_to_element, NumpyEncoder
from flare.utils.learner import (
    subset_of_frame_by_element,
    is_std_in_bound_per_species,
    is_force_in_bound_per_species,
)
from flare.mgp import MappedGaussianProcess
from flare.parameters import Parameters


class TrajectoryTrainer:
    def __init__(
        self,
        frames: List[Structure],
        gp: Union[GaussianProcess, MappedGaussianProcess],
        rel_std_tolerance: float = 4,
        abs_std_tolerance: float = 1,
        abs_force_tolerance: float = 0,
        max_force_error: float = inf,
        parallel: bool = False,
        n_cpus: int = 1,
        skip: int = 1,
        validate_ratio: float = 0.0,
        calculate_energy: bool = False,
        output_name: str = "gp_from_aimd",
        print_as_xyz: bool = False,
        pre_train_max_iter: int = 50,
        max_atoms_from_frame: int = np.inf,
        max_trains: int = np.inf,
        min_atoms_per_train: int = 1,
        shuffle_frames: bool = False,
        verbose: str = "INFO",
        pre_train_on_skips: int = -1,
        pre_train_seed_frames: List[Structure] = None,
        pre_train_seed_envs: List[Tuple[AtomicEnvironment, "np.array"]] = None,
        pre_train_atoms_per_element: dict = None,
        train_atoms_per_element: dict = None,
        predict_atoms_per_element: dict = None,
        train_checkpoint_interval: int = 1,
        checkpoint_interval: int = 1,
        atom_checkpoint_interval: int = 100,
        print_training_plan: bool = True,
        model_format: str = "json",
    ):
        """
        Class which trains a GP off of an AIMD trajectory, and generates
        error statistics between the DFT and GP calls.

        There are a variety of options which can give you a finer control
        over the training process.

        :param frames: List of structures to evaluate / train GP on
        :param gp: Gaussian Process object
        :param rel_std_tolerance: Train if uncertainty is above this *
            noise variance hyperparameter
        :param abs_std_tolerance: Train if uncertainty is above this
        :param abs_force_tolerance: Add atom force error exceeds this
        :param max_force_error: Don't add atom if force error exceeds this
        :param parallel: Use parallel functions or not
        :param validate_ratio: Fraction of frames used for validation
        :param skip: Skip through frames
        :param calculate_energy: Use local energy kernel or not
        :param output_name: Write output of training to this file
        :param print_as_xyz: If True, print the configurations in xyz format
        :param max_atoms_from_frame: Largest # of atoms added from one frame
        :param min_atoms_per_train: Only train when this many atoms have been
            added
        :param max_trains: Stop training GP after this many calls to train
        :param n_cpus: Number of CPUs to parallelize over for parallelization
                over atoms
        :param shuffle_frames: Randomize order of frames for better training
        :param verbose: same as logging level, "WARNING", "INFO", "DEBUG"
        :param pre_train_on_skips: Train model on every n frames before running
        :param pre_train_seed_frames: Frames to train on before running
        :param pre_train_seed_envs: Environments to train on before running
        :param pre_train_atoms_per_element: Max # of environments to add from
            each species in the seed pre-training steps
        :param train_atoms_per_element: Max # of environments to add from
            each species in the training steps
        :param predict_atoms_per_element: Choose a random subset of N random
            atoms from each specified element to predict on. For instance,
            {"H":5} will only predict the forces and uncertainties
            associated with 5 Hydrogen atoms per frame. Elements not
            specified will be predicted as normal. This is useful for
            systems where you are most interested in a subset of elements.
            This will result in a faster but less exhaustive learning process.
        :param checkpoint_interval: Will be deprecated. Same as
                            train_checkpoint_interval
        :param train_checkpoint_interval: How often to write model after
                        trainings
        :param atom_checkpoint_interval: How often to write model after atoms are
            added (since atoms may be added without training)
        :param model_format: Format to write GP model to
        :param print_training_plan: Write which atoms in which frames that
            triggered uncertainty or force conditions, so that training can
            be 'fast-forwarded' later. Also useful for gauging MGP results and
            then applying the atoms with high uncertainty and error to a GP.
        """

        # Set up parameters
        self.frames = frames
        if shuffle_frames:
            np.random.shuffle(frames)
            if print_training_plan:
                warnings.warn(
                    "Frames are shuffled so training plan will not"
                    " map onto the structures used; Try to "
                    "shuffle the frames outside of the GPFA module "
                    "for now."
                )

        # GP Training and Execution parameters
        self.gp = gp
        # Check to see if GP is MGP for later flagging
        self.mgp = isinstance(gp, MappedGaussianProcess)
        self.rel_std_tolerance = rel_std_tolerance
        self.abs_std_tolerance = abs_std_tolerance
        self.abs_force_tolerance = abs_force_tolerance
        self.max_force_error = max_force_error
        self.max_trains = max_trains
        self.max_atoms_from_frame = max_atoms_from_frame
        self.min_atoms_per_train = min_atoms_per_train
        self.predict_atoms_per_element = predict_atoms_per_element
        self.train_count = 0
        self.calculate_energy = calculate_energy
        self.n_cpus = n_cpus

        if parallel is True:
            warnings.warn(
                "Parallel flag will be deprecated;we will instead use n_cpu alone.",
                DeprecationWarning,
            )

        # Set prediction function based on if forces or energies are
        # desired, and parallelization accordingly
        if not self.mgp:
            if calculate_energy:
                self.pred_func = predict_on_structure_par_en
            else:
                self.pred_func = predict_on_structure_par

        elif self.mgp:
            self.pred_func = predict_on_structure_mgp

        # Parameters for negotiating with the training frames

        # To later be filled in using the time library
        self.start_time = None

        self.skip = skip
        assert (
            isinstance(skip, int) and skip >= 1
        ), "Skip needs to be a positive integer."
        self.validate_ratio = validate_ratio
        assert 0 <= validate_ratio <= 1, "validate_ratio needs to be [0,1]"

        # Set up for pretraining
        self.pre_train_max_iter = pre_train_max_iter
        self.pre_train_on_skips = pre_train_on_skips
        self.seed_envs = [] if pre_train_seed_envs is None else pre_train_seed_envs
        self.seed_frames = (
            [] if pre_train_seed_frames is None else pre_train_seed_frames
        )

        self.pre_train_env_per_species = (
            {} if pre_train_atoms_per_element is None else pre_train_atoms_per_element
        )
        self.train_env_per_species = (
            {} if train_atoms_per_element is None else train_atoms_per_element
        )

        # Convert to Coded Species
        if self.pre_train_env_per_species:
            pre_train_species = list(self.pre_train_env_per_species.keys())
            for key in pre_train_species:
                self.pre_train_env_per_species[
                    element_to_Z(key)
                ] = self.pre_train_env_per_species[key]

        # Output parameters
        self.output = Output(
            output_name, verbose, print_as_xyz=print_as_xyz, always_flush=True
        )
        self.logger_name = self.output.basename + "log"
        self.train_checkpoint_interval = (
            train_checkpoint_interval or checkpoint_interval
        )
        self.atom_checkpoint_interval = atom_checkpoint_interval

        self.model_format = model_format
        self.output_name = output_name
        self.print_training_plan = print_training_plan

        # Defining variables to be used later
        self.curr_step = 0
        self.train_count = 0
        self.start_time = time.time()

    def pre_run(self):
        """
        Various tasks to set up the AIMD training before commencing
        the run through the AIMD trajectory.
        1. Print the output.
        2. Pre-train the GP with the seed frames and
        environments. If no seed frames or environments and the GP has no
        training set, then seed with at least one atom from each
        """

        if self.mgp:
            raise NotImplementedError("Pre-running notyet configured for MGP")
        self.output.write_header(
            str(self.gp),
            dt=0,
            Nsteps=len(self.frames),
            structure=None,
            std_tolerance=(self.rel_std_tolerance, self.abs_std_tolerance),
            optional={
                "GP Statistics": json.dumps(self.gp.training_statistics),
                "GP Name": self.gp.name,
                "GP Write Name": self.output_name + "_model." + self.model_format,
            },
        )

        self.start_time = time.time()
        logger = logging.getLogger(self.logger_name)
        logger.debug("Now beginning pre-run activity.")
        # If seed environments were passed in, add them to the GP.

        for point in self.seed_envs:
            self.gp.add_one_env(point[0], point[1], train=False)

        # No training set ("blank slate" run) and no seeds specified:
        # Take one of each atom species in the first frame
        # so all atomic species are represented in the first step.
        # Otherwise use the seed frames passed in by user.

        # Remove frames used as seed from later part of training
        if self.pre_train_on_skips > 0:
            self.seed_frames = []
            newframes = []
            for i in range(len(self.frames)):
                if (i % self.pre_train_on_skips) == 0:
                    self.seed_frames += [self.frames[i]]
                else:
                    newframes += [self.frames[i]]
            self.frames = newframes

        # If the GP is empty, use the first frame as a seed frame.
        elif len(self.gp.training_data) == 0 and len(self.seed_frames) == 0:
            self.seed_frames = [self.frames[0]]
            self.frames = self.frames[1:]

        atom_count = 0
        for frame in self.seed_frames:
            train_atoms = []
            for species_i in set(frame.coded_species):
                # Get a randomized set of atoms of species i from the frame
                # So that it is not always the lowest-indexed atoms chosen
                atoms_of_specie = frame.indices_of_specie(species_i)
                np.random.shuffle(atoms_of_specie)
                n_at = len(atoms_of_specie)
                # Determine how many to add based on user defined cutoffs
                n_to_add = min(
                    n_at,
                    self.pre_train_env_per_species.get(species_i, inf),
                    self.max_atoms_from_frame,
                )

                for atom in atoms_of_specie[:n_to_add]:
                    train_atoms.append(atom)
                    atom_count += 1

            self.update_gp_and_print(
                frame=frame,
                train_atoms=train_atoms,
                uncertainties=[],
                train=False,
            )

        logger = logging.getLogger(self.logger_name)
        if atom_count > 0:
            logger.info(
                f"Added {atom_count} atoms to "
                "pretrain.\n"
                "Pre-run GP Statistics: "
                f"{json.dumps(self.gp.training_statistics)} "
            )

        if (self.seed_envs or atom_count or self.seed_frames) and (
            self.pre_train_max_iter or self.max_trains
        ):
            logger.debug(
                "Now commencing pre-run training of GP (which has "
                "non-empty training set)"
            )
            time0 = time.time()
            self.train_gp(max_iter=self.pre_train_max_iter)
            logger.debug(f"Done train_gp {time.time()-time0}")
        else:
            logger.debug(
                "Now commencing pre-run set up of GP (which has non-empty training set)"
            )
            time0 = time.time()
            self.gp.check_L_alpha()
            logger.debug(f"Done check_L_alpha {time.time()-time0}")

        if self.model_format and not self.mgp:
            self.gp.write_model(f"{self.output_name}_prerun", self.model_format)

    def run(self):
        """
        Loop through frames and record the error between
        the GP predictions and the ground-truth forces. Train the GP and update
        the training set upon the triggering of the uncertainty or force error
        threshold.

        :return: None
        """

        # Perform pre-run, in which seed trames are used.
        logger = logging.getLogger(self.logger_name)
        logger.debug("Commencing run with pre-run...")
        if not self.mgp:
            self.pre_run()

        # Past this frame, stop adding atoms to the training set
        #  (used for validation of model)
        train_frame = int(len(self.frames[:: self.skip]) * (1 - self.validate_ratio))

        # Loop through trajectory.
        cur_atoms_added_train = 0  # Track atoms added for training
        cur_atoms_added_write = 0  # Track atoms added for writing
        cur_trains_done_write = 0  # Track training done for writing

        # Keep track of which atoms trigger force / uncertainty condition
        training_plan = {}

        for i, cur_frame in enumerate(self.frames[:: self.skip]):

            frame_start_time = time.time()
            logger.info(f"=====NOW ON FRAME {i}=====")

            # If no predict_atoms_per_element was specified, predict_atoms
            # will be equal to every atom in the frame.
            predict_atoms = subset_of_frame_by_element(
                cur_frame, self.predict_atoms_per_element
            )

            # Atoms which are skipped will have NaN as their force / std values
            local_energies = None

            # Three different predictions: Either MGP, GP with energy,
            # or GP without
            if self.mgp:
                pred_forces, pred_stds, local_energies = self.pred_func(
                    structure=cur_frame,
                    mgp=self.gp,
                    write_to_structure=False,
                    selective_atoms=predict_atoms,
                    skipped_atom_value=np.nan,
                    energy=True,
                )
            elif self.calculate_energy:
                pred_forces, pred_stds, local_energies = self.pred_func(
                    structure=cur_frame,
                    gp=self.gp,
                    n_cpus=self.n_cpus,
                    write_to_structure=False,
                    selective_atoms=predict_atoms,
                    skipped_atom_value=np.nan,
                )
            else:
                pred_forces, pred_stds = self.pred_func(
                    structure=cur_frame,
                    gp=self.gp,
                    n_cpus=self.n_cpus,
                    write_to_structure=False,
                    selective_atoms=predict_atoms,
                    skipped_atom_value=np.nan,
                )

            # Get Error
            dft_forces = cur_frame.forces
            dft_energy = cur_frame.energy
            error = np.abs(pred_forces - dft_forces)

            # Create dummy frame with the predicted forces written
            dummy_frame = deepcopy(cur_frame)
            dummy_frame.forces = pred_forces
            dummy_frame.stds = pred_stds
            cur_frame.stds = pred_stds

            self.output.write_gp_dft_comparison(
                curr_step=i,
                frame=dummy_frame,
                start_time=frame_start_time,
                dft_forces=dft_forces,
                dft_energy=dft_energy,
                error=error,
                local_energies=local_energies,
                KE=0,
                cell=cur_frame.cell,
            )

            logger.debug(
                f"Single frame calculation time {time.time()-frame_start_time}"
            )

            if i < train_frame:
                # Noise hyperparameter & relative std tolerance is not for mgp.
                if self.mgp:
                    noise = 0
                else:
                    noise = Parameters.get_noise(
                        self.gp.hyps_mask, self.gp.hyps, constraint=False
                    )
                std_in_bound, std_train_atoms = is_std_in_bound_per_species(
                    rel_std_tolerance=self.rel_std_tolerance,
                    abs_std_tolerance=self.abs_std_tolerance,
                    noise=noise,
                    structure=dummy_frame,
                    max_atoms_added=self.max_atoms_from_frame,
                    max_by_species=self.train_env_per_species,
                )

                # Get max force error atoms
                force_in_bound, force_train_atoms = is_force_in_bound_per_species(
                    abs_force_tolerance=self.abs_force_tolerance,
                    predicted_forces=pred_forces,
                    label_forces=dft_forces,
                    structure=dummy_frame,
                    max_atoms_added=self.max_atoms_from_frame,
                    max_by_species=self.train_env_per_species,
                    max_force_error=self.max_force_error,
                )

                if not std_in_bound or not force_in_bound:

                    # -1 is returned from the is_in_bound methods,
                    # so filter that out and the use sets to remove repeats
                    train_atoms = list(
                        set(std_train_atoms).union(force_train_atoms) - {-1}
                    )

                    # Record frame and training atoms, uncertainty, error
                    force_errors = list(np.abs(pred_forces - dft_forces))
                    uncertainties = list(dummy_frame.stds)
                    training_plan[int(i)] = [
                        (int(a), uncertainties[a], force_errors[a]) for a in train_atoms
                    ]

                    # Compute mae and write to output;
                    # Add max uncertainty atoms to training set
                    self.update_gp_and_print(
                        cur_frame,
                        train_atoms=train_atoms,
                        uncertainties=pred_stds[train_atoms],
                        train=False,
                    )
                    cur_atoms_added_train += len(train_atoms)
                    cur_atoms_added_write += len(train_atoms)
                    # Re-train if number of sampled atoms is high enough

                    if (
                        cur_atoms_added_train >= self.min_atoms_per_train
                        or (i + 1) == train_frame
                    ):
                        if self.train_count < self.max_trains:
                            self.train_gp()
                            cur_trains_done_write += 1
                        else:
                            self.gp.update_L_alpha()
                        cur_atoms_added_train = 0
                    else:
                        self.gp.update_L_alpha()

                    # Loop to decide of a model should be written this
                    # iteration
                    will_write = False

                    if (
                        self.train_checkpoint_interval
                        and cur_trains_done_write
                        and self.train_checkpoint_interval <= cur_trains_done_write
                    ):
                        will_write = True
                        cur_trains_done_write = 0

                    if (
                        self.atom_checkpoint_interval
                        and cur_atoms_added_write
                        and self.atom_checkpoint_interval <= cur_atoms_added_write
                    ):
                        will_write = True
                        cur_atoms_added_write = 0

                    if self.model_format and will_write:
                        self.gp.write_model(
                            f"{self.output_name}_checkpt", self.model_format
                        )

                if (i + 1) == train_frame and not self.mgp:
                    self.gp.check_L_alpha()

        # Print training statistics for GP model used
        conclusion_strings = []
        conclusion_strings.append(
            "Final GP statistics:" + json.dumps(self.gp.training_statistics)
        )
        self.output.conclude_run(conclusion_strings)

        if self.print_training_plan:
            with open(f"{self.output_name}_training_plan.json", "w") as f:
                f.write(json.dumps(training_plan, cls=NumpyEncoder))

        if self.model_format and not self.mgp:
            self.gp.write_model(f"{self.output_name}_model", self.model_format)

    def update_gp_and_print(
        self,
        frame: Structure,
        train_atoms: List[int],
        uncertainties: List[int] = None,
        train: bool = True,
    ):
        """
        Update the internal GP model training set with a list of training
        atoms indexing atoms within the frame. If train is True, re-train
        the GP by optimizing hyperparameters.
        :param frame: Structure to train on
        :param train_atoms: Index atoms to train on
        :param uncertainties: Uncertainties to print, pass in [] to silence
        :param train: Train or not
        :return: None
        """

        # Group added atoms by species for easier output
        added_species = [Z_to_element(frame.coded_species[at]) for at in train_atoms]
        added_atoms = {spec: [] for spec in set(added_species)}

        for atom, spec in zip(train_atoms, added_species):
            added_atoms[spec].append(atom)

        logger = logging.getLogger(self.logger_name)
        logger.info(
            "Adding atom(s) "
            f"{json.dumps(added_atoms,cls=NumpyEncoder)}"
            " to the training set."
        )

        if uncertainties is None:
            uncertainties = frame.stds[train_atoms]

        if uncertainties is not None and len(uncertainties) != 0:
            logger.info(f"Uncertainties: {uncertainties}.")

        logger.info(f"New GP Statistics: {json.dumps(self.gp.training_statistics)}\n")

        # update gp model; handling differently if it's an MGP
        if not self.mgp:
            self.gp.update_db(frame, frame.forces, custom_range=train_atoms)

            if train:
                self.train_gp()

        else:
            logger.warning("Warning: Adding data to an MGP is not yet supported.")

    def train_gp(self, max_iter: int = None):
        """
        Train the Gaussian process and write the results to the output file.

        :param max_iter: Maximum iterations associated with this training run,
            overriding the Gaussian Process's internally set maxiter.
        :type max_iter: int
        """
        logger = logging.getLogger(self.logger_name)

        if self.mgp:
            logger.debug("Training skipped because of MGP")
            return

        logger.debug("Train GP")

        logger_train = self.output.basename + "hyps"

        # TODO: Improve flexibility in GP training to make this next step
        # unnecessary, so maxiter can be passed as an argument

        # Don't train if maxiter == 0
        if max_iter == 0:
            self.gp.check_L_alpha()
        elif max_iter is not None:
            temp_maxiter = self.gp.maxiter
            self.gp.maxiter = max_iter
            self.gp.train(logger_name=logger_train)
            self.gp.maxiter = temp_maxiter
        else:
            self.gp.train(logger_name=logger_train)

        hyps, labels = Parameters.get_hyps(
            self.gp.hyps_mask, self.gp.hyps, constraint=False, label=True
        )
        if labels is None:
            labels = self.gp.hyp_labels
        self.output.write_hyps(
            labels,
            hyps,
            self.start_time,
            self.gp.likelihood,
            self.gp.likelihood_gradient,
            hyps_mask=self.gp.hyps_mask,
        )
        self.train_count += 1


def parse_frame_block(chunk: str, compute_errors: bool = True):
    # i loops through individual atom's info

    frame_atoms = []
    frame_positions = []
    gp_forces = []
    dft_forces = []
    stds = []
    added_atoms = {}
    frame_species_maes = {}

    # First two lines contain frame # and header
    for i in range(2, len(chunk)):

        # Lines with data will be long; stop when at end of atom data
        # Terminate at blank line between results
        if chunk[i] != "\n":
            split = chunk[i].split()

            frame_atoms.append(split[0])

            frame_positions.append([float(split[1]), float(split[2]), float(split[3])])
            gp_forces.append([float(split[4]), float(split[5]), float(split[6])])
            stds.append([float(split[7]), float(split[8]), float(split[9])])

            dft_forces.append([float(split[10]), float(split[11]), float(split[12])])
        else:
            break

    # Loop through information in frame after Data
    cell = None
    for i in range(len(frame_positions) + 2, len(chunk)):
        if "cell" in chunk[i]:
            split_line = chunk[i].strip().split(":")
            cell = np.array(json.loads(split_line[1]))

        if "Adding atom(s)" in chunk[i]:
            # Splitting to target the 'added atoms' substring
            split_line = chunk[i][15:-21]
            added_atoms = json.loads(split_line.strip())

        if "type " in chunk[i]:
            cur_line = chunk[i].split()
            frame_species_maes[cur_line[1]] = float(cur_line[3])

    cur_frame_stats = {
        "species": frame_atoms,
        "positions": np.array(frame_positions),
        "gp_forces": np.array(gp_forces),
        "dft_forces": np.array(dft_forces),
        "gp_stds": np.array(stds),
        "added_atoms": added_atoms,
        "maes_by_species": frame_species_maes,
    }
    if compute_errors:
        cur_frame_stats["force_errors"] = np.array(gp_forces) - np.array(dft_forces)
    if cell is not None:
        cur_frame_stats["cell"] = cell

    return cur_frame_stats


def parse_trajectory_trainer_output(
    file: str, return_gp_data: bool = False, compute_errors: bool = True
) -> Union[List[dict], Tuple[List[dict], dict]]:
    """
    Reads output of a TrajectoryTrainer run by frame. return_gp_data returns
    data about GP model growth useful for visualizing progress of model
    training.

    :param file: filename of output
    :param return_gp_data: flag for returning extra GP data
    :param compute_errors: Compute deviation from GP and DFT forces.
    :return: List of dictionaries with keys 'species', 'positions',
        'gp_forces', 'dft_forces', 'gp_stds', 'added_atoms', and
        'maes_by_species', optionally, gp_data dictionary
    """

    with open(file, "r") as f:
        lines = f.readlines()
        num_lines = len(lines)

    # Get indexes where frames begin, and include the index of the final line
    frame_indexes = [i for i in range(num_lines) if "-Frame:" in lines[i]] + [num_lines]

    frames = []

    # Central parsing loop
    for n in range(len(frame_indexes) - 1):
        # Start at +2 to skip frame marker and header of table of data
        # Set up values for current frame which will be populated

        range(frame_indexes[n] + 2, frame_indexes[n + 1])

        near_index = frame_indexes[n]
        far_index = frame_indexes[n + 1]

        frames.append(parse_frame_block(lines[near_index:far_index]))

    if not return_gp_data:
        return frames

    # Compute information about GP training
    # to study GP growth and performance over trajectory

    gp_stats_line = [
        line for line in lines[:30] if "GP Statistics" in line and "Pre-run" not in line
    ][0][15:].strip()

    initial_gp_statistics = json.loads(gp_stats_line)

    # Get pre_run statistics (if pre-run was done):
    pre_run_gp_statistics = None
    pre_run_gp_stats_line = [line for line in lines if "Pre-run GP" in line]
    if pre_run_gp_stats_line:
        pre_run_gp_statistics = json.loads(pre_run_gp_stats_line[0][22:].strip())

    # Compute cumulative GP size
    cumulative_gp_size = [int(initial_gp_statistics["N"])]

    if pre_run_gp_stats_line:
        cumulative_gp_size.append(int(pre_run_gp_statistics["N"]))

    running_total = cumulative_gp_size[-1]

    for frame in frames:

        added_atom_dict = frame["added_atoms"]
        for val in added_atom_dict.values():
            running_total += len(val)
        cumulative_gp_size.append(running_total)

    # Compute MAEs for each element over time
    all_species = set()
    for frame in frames:
        all_species = all_species.union(set(frame["species"]))

    all_species = list(all_species)
    mae_by_elt = {elt: [] for elt in all_species}

    for frame in frames:
        for elt in all_species:
            cur_mae = frame["maes_by_species"].get(elt, np.nan)
            mae_by_elt[elt].append(cur_mae)

    gp_data = {
        "init_stats": initial_gp_statistics,
        "pre_train_stats": pre_run_gp_statistics,
        "cumulative_gp_size": cumulative_gp_size,
        "mae_by_elt": mae_by_elt,
    }

    return frames, gp_data


def structures_from_gpfa_output(frame_dictionaries: List[dict]) -> List[Structure]:
    """
    Takes as input the first output from the `parse_trajectory_trainer_output`
    function and turns it into a series of FLARE structures, with DFT forces mapped
    onto the structures.

    :param frame_dictionaries: The list of dictionaries which describe each GPFA frame.
    :return:
    """

    structures = []
    for frame in frame_dictionaries:
        structures.append(
            Structure(
                cell=frame["cell"],
                species=frame["species"],
                positions=frame["positions"],
                forces=frame["dft_forces"],
            )
        )
    return structures



"""Cubic spline functions used for interpolation. 
"""
import numpy as np
import numpy
from flare.mgp.cubic_splines_numba import *


class PCASplines:
    """
    Build splines for PCA decomposition, mainly used for the mapping of the variance

    :param l_bounds: lower bound for the interpolation. \
        E.g. 1-d for two-body, 3-d for three-body.
    :type l_bounds: numpy array
    :param u_bounds: upper bound for the interpolation.
    :type u_bounds: numpy array
    :param orders: grid numbers in each dimension. E.g, 1-d for two-body, \
        3-d for three-body, should be positive integers.
    :type orders: numpy array
    :param svd_rank: rank for decomposition of variance matrix,\
        also equal to the number of mappings constructed for mapping variance.\
        For two-body `svd_rank<=min(grid_num, train_size*3)`, \
        for three-body `svd_rank<=min(grid_num_in_cube, train_size*3)`
    :type svd_rank: int
    """

    def __init__(self, l_bounds, u_bounds, orders, svd_rank):
        self.svd_rank = svd_rank
        self.models = []
        for r in range(svd_rank):
            spline_u = CubicSpline(l_bounds, u_bounds, orders)
            self.models.append(spline_u)

    def build_cubic(self, y, u_bounds, l_bounds, orders):
        dim_0 = 1
        for d in range(len(y.shape) - 1):
            dim_0 *= y.shape[d]
        dim_1 = y.shape[-1]

        var_matr = np.reshape(y, (dim_0, dim_1))
        models = []
        for r in range(self.svd_rank):
            spline_u = CubicSpline(l_bounds, u_bounds, orders, var_matr[:, r])
            models.append(spline_u)
        return models

    def set_values(self, y):
        dim_0 = 1
        for d in range(len(y.shape) - 1):
            dim_0 *= y.shape[d]
        dim_1 = y.shape[-1]

        var_matr = np.reshape(y, (dim_0, dim_1))
        U, S, Vh = np.linalg.svd(var_matr, full_matrices=False)
        self.V = Vh[: self.svd_rank, :].T
        for r in range(self.svd_rank):
            self.models[r].set_values(S[r] * U[:, r])

    def __call__(self, x):
        y_pred = []
        rank = self.svd_rank
        for r in range(rank):
            y_pred.append(self.models[r](x))
        return np.array(y_pred)


class CubicSpline:

    """
    Forked from Github repository: https://github.com/EconForge/interpolation.py.\
    High-level API for cubic splines. \
    Class representing a cubic spline interpolator on a regular cartesian grid.

    Creates a cubic spline interpolator on a regular cartesian grid.

    Args:
        a (numpy array of size d (float)): Lower bounds of the cartesian grid.
        b (numpy array of size d (float)): Upper bounds of the cartesian grid.
        orders (numpy array of size d (int)): Number of nodes along each \
            dimension (=(n1,...,nd) )

    Other Parameters:
        values (numpy array (float)): (optional, (n1 x ... x nd) array). \
            Values on the nodes of the function to interpolate.
    """

    __grid__ = None
    __values__ = None
    __coeffs__ = None

    def __init__(self, a, b, orders, values=None):

        self.d = len(a)
        assert len(b) == self.d
        assert len(orders) == self.d
        self.a = np.array(a, dtype=float)
        self.b = np.array(b, dtype=float)
        self.orders = np.array(orders, dtype=int)
        self.dtype = self.a.dtype
        self.__coeffs__ = None

        if values is not None:
            self.set_values(values)

    def set_values(self, values):
        """Set values on the nodes for the function to interpolate."""

        values = np.array(values, dtype=float)

        if not np.all(np.isfinite(values)):
            raise Exception("Trying to interpolate non-finite values")

        sh = self.orders.tolist()
        sh2 = [e + 2 for e in self.orders]

        values = values.reshape(sh)

        self.__values__ = values

        # this should be done without temporary memory allocation
        self.__coeffs__ = filter_coeffs(self.a, self.b, self.orders, values)

    def interpolate(self, points, values=None, with_derivatives=False):
        """
        Interpolate spline at a list of points.

        :param points: (array-like) list of point where the spline is evaluated.
        :param values: (optional) container for inplace computation.
        :return values: (array-like) list of point where the spline is evaluated.
        """

        if not np.all(np.isfinite(points)):
            raise Exception("Spline interpolator evaluated at non-finite points.")

        if not with_derivatives:
            if points.ndim == 1:
                # evaluate only on one point
                points = np.array([points])
            N, d = points.shape
            assert d == self.d
            if values is None:
                values = np.empty(N, dtype=self.dtype)
            vec_eval_cubic_spline(
                self.a, self.b, self.orders, self.__coeffs__, points, values
            )
            return values
        else:
            N, d = points.shape
            assert d == self.d
            values, dvalues = vec_eval_cubic_splines_G(
                self.a,
                self.b,
                self.orders,
                self.__coeffs__,
                points,
                values,
                dvalues=None,
            )

            return values, dvalues

    @property
    def grid(self):
        """Cartesian enumeration of all nodes."""

        if self.__grid__ is None:
            self.__grid__ = mlinspace(self.a, self.b, self.orders)
        return self.__grid__

    def __call__(self, s, with_derivatives=False):
        """Interpolate the spline at one or many points"""

        if s.ndim == 1:
            res = self.__call__(numpy.atleast_2d(s))
            return res[0]

        return self.interpolate(s, with_derivatives=with_derivatives)


def vec_eval_cubic_spline(a, b, orders, coefs, points, values=None):
    """
    Forked from Github repository: https://github.com/EconForge/interpolation.py.\
    Evaluates a cubic spline at many points

    :param a: Lower bounds of the cartesian grid.
    :type  a: numpy array of size d (float)
    :param b: Upper bounds of the cartesian grid.
    :type  b: numpy array of size d (float)
    :param orders: Number of nodes along each dimension (=(n1,...,nd) )
    :type  orders: numpy array of size d (int)
    :param coefs: Filtered coefficients.
    :type  coefs: array of dimension d, and size (n1+2, ..., nd+2)
    :param point: List of points where the splines must be interpolated.
    :type  point: array of size N x d
    :param values: (optional) If not None, contains the result.
    :type  values: array of size N

    :return value: Interpolated values. values[i] contains spline evaluated at point points[i,:].
    :type   value: array of size N
    """

    a = numpy.array(a, dtype=float)
    b = numpy.array(b, dtype=float)
    orders = numpy.array(orders, dtype=int)

    d = a.shape[0]

    if values is None:
        N = points.shape[0]
        values = numpy.empty(N)

    if d == 1:
        vec_eval_cubic_spline_1(a, b, orders, coefs, points, values)
    elif d == 2:
        vec_eval_cubic_spline_2(a, b, orders, coefs, points, values)
    elif d == 3:
        vec_eval_cubic_spline_3(a, b, orders, coefs, points, values)
    elif d == 4:
        vec_eval_cubic_spline_4(a, b, orders, coefs, points, values)

    return values


def vec_eval_cubic_splines_G(a, b, orders, mcoefs, points, values=None, dvalues=None):

    a = numpy.array(a, dtype=float)
    b = numpy.array(b, dtype=float)
    orders = numpy.array(orders, dtype=int)

    d = a.shape[0]
    N = points.shape[0]
    # n_sp = mcoefs.shape[-1]
    n_sp = 1

    if values is None:
        values = numpy.empty((N, n_sp))

    if dvalues is None:
        dvalues = numpy.empty((N, d, n_sp))

    if d == 1:
        vec_eval_cubic_splines_G_1(a, b, orders, mcoefs, points, values, dvalues)

    elif d == 2:
        vec_eval_cubic_splines_G_2(a, b, orders, mcoefs, points, values, dvalues)

    elif d == 3:
        vec_eval_cubic_splines_G_3(a, b, orders, mcoefs, points, values, dvalues)

    elif d == 4:
        vec_eval_cubic_splines_G_4(a, b, orders, mcoefs, points, values, dvalues)

    return [values, dvalues]



import logging
import math
import multiprocessing as mp
import numpy as np
import time

from typing import List, Callable
from flare.kernels.utils import from_mask_to_args, from_grad_to_mask

_global_training_data = {}
_global_training_labels = {}
_global_training_structures = {}
_global_energy_labels = {}


def queue_wrapper(result_queue, wid, func, args):
    """
    wrapper function for multiprocessing queue
    """
    result_queue.put((wid, func(*args)))


def partition_matrix(n_sample, size, n_cpus):
    """
    partition the training data for matrix calculation
    the number of blocks are close to n_cpus
    since mp.Process does not allow to change the thread number
    """

    # divide the block by n_cpu partitions, with size n_sample0
    # if the divided chunk is smaller than the requested chunk n_sample
    # use the requested chunk size
    n_unique_elements = size * (size + 1) / 2
    n_sample0 = int(math.ceil(np.sqrt(n_unique_elements / n_cpus)))
    if n_sample0 > n_sample:
        n_sample = n_sample0

    block_id = []
    nbatch = 0
    nbatch1 = 0
    nbatch2 = 0
    e1 = 0
    while e1 < size:
        s1 = int(n_sample * nbatch1)
        e1 = int(np.min([s1 + n_sample, size]))
        nbatch2 = nbatch1
        nbatch1 += 1
        e2 = 0
        while e2 < size:
            s2 = int(n_sample * nbatch2)
            e2 = int(np.min([s2 + n_sample, size]))
            block_id += [(s1, e1, s2, e2)]
            nbatch2 += 1
            nbatch += 1

    return block_id, nbatch


def partition_matrix_custom(n_sample: int, start1, end1, start2, end2, n_cpus):
    """
    Partition a specified portion of a matrix.
    """

    n_rows = end1 - start1
    n_cols = end2 - start2
    n_unique_elements = n_rows * n_cols
    n_sample0 = int(math.ceil(np.sqrt(n_unique_elements / n_cpus)))
    if n_sample0 > n_sample:
        n_sample = n_sample0

    block_id = []
    n_block = 0  # total number of blocks
    nbatch1 = 0
    nbatch2 = 0
    e1 = 0
    while e1 < end1:
        s1 = start1 + n_sample * nbatch1
        e1 = np.min([s1 + n_sample, end1])
        nbatch1 += 1

        e2 = 0
        nbatch2 = 0
        while e2 < end2:
            s2 = start2 + n_sample * nbatch2
            e2 = np.min([s2 + n_sample, end2])
            block_id += [(s1, e1, s2, e2)]
            nbatch2 += 1
            n_block += 1  # update block count

    return block_id, n_block


def partition_vector(n_sample, size, n_cpus):
    """
    partition the training data for vector calculation
    the number of blocks are the same as n_cpus
    since mp.Process does not allow to change the thread number
    """
    n_sample0 = int(math.ceil(size / n_cpus))
    if n_sample0 > n_sample:
        n_sample = n_sample0

    block_id = []
    nbatch = 0
    e = 0
    while e < size:
        s = n_sample * nbatch
        e = np.min([s + n_sample, size])
        block_id += [(s, e)]
        nbatch += 1
    return block_id, nbatch


def partition_force_energy_block(n_sample: int, size1: int, size2: int, n_cpus: int):
    """Special partition method for the force/energy block. Because the number
    of environments in a structure can vary, we only split up the environment
    list, which has length size1.

    Note that two sizes need to be specified: the size of the envionment
    list and the size of the structure list.

    Args:
        n_sample (int): Number of environments per processor.
        size1 (int): Size of the environment list.
        size2 (int): Size of the structure list.
        n_cpus (int): Number of cpus.
    """

    n_sample0 = int(math.ceil(size1 / n_cpus))
    if n_sample0 > n_sample:
        n_sample = n_sample0

    block_id = []
    nbatch = 0
    e = 0

    while e < size1:
        s = n_sample * nbatch
        e = np.min([s + n_sample, size1])
        block_id += [(s, e, 0, size2)]
        nbatch += 1

    return block_id, nbatch


def partition_update(n_sample, size, old_size, n_cpus):

    n_unique_elements = (size - old_size) * size
    n_sample0 = int(math.ceil(np.sqrt(n_unique_elements / n_cpus)))
    if n_sample0 > n_sample:
        n_sample = n_sample0

    ns_new = int(math.ceil((size - old_size) / n_sample))
    old_ns = int(math.ceil(old_size / n_sample))

    nbatch = 0
    block_id = []
    for ibatch1 in range(old_ns):
        s1 = int(n_sample * ibatch1)
        e1 = int(np.min([s1 + n_sample, old_size]))

        for ibatch2 in range(ns_new):
            s2 = int(n_sample * ibatch2) + old_size
            e2 = int(np.min([s2 + n_sample, size]))
            block_id += [(s1, e1, s2, e2)]
            nbatch += 1

    for ibatch1 in range(ns_new):
        s1 = int(n_sample * ibatch1) + old_size
        e1 = int(np.min([s1 + n_sample, size]))

        for ibatch2 in range(ns_new):
            s2 = int(n_sample * ibatch2) + old_size
            e2 = int(np.min([s2 + n_sample, size]))
            block_id += [(s1, e1, s2, e2)]
            nbatch += 1

    return block_id, nbatch


def obtain_noise_len(hyps, hyps_mask):
    """
    obtain the noise parameter from hyps and mask
    """

    # assume sigma_n is the final hyperparameter
    sigma_n = hyps[-1]
    # correct it if map is defined
    non_noise_hyps = len(hyps) - 1
    train_noise = True
    if hyps_mask is not None:
        train_noise = hyps_mask.get("train_noise", True)
        if train_noise is False:
            sigma_n = hyps_mask["original_hyps"][-1]
            non_noise_hyps = len(hyps)

    return sigma_n, non_noise_hyps, train_noise


# --------------------------------------------------------------------------
#                   Parallel matrix/vector construction
# --------------------------------------------------------------------------


def parallel_matrix_construction(
    pack_function,
    hyps,
    name,
    kernel,
    cutoffs,
    hyps_mask,
    block_id,
    nbatch,
    size1,
    size2,
    m1,
    m2,
    symm=True,
):
    result_queue = mp.Queue()
    children = []
    for wid in range(nbatch):
        s1, e1, s2, e2 = block_id[wid]
        children.append(
            mp.Process(
                target=queue_wrapper,
                args=(
                    result_queue,
                    wid,
                    pack_function,
                    (hyps, name, s1, e1, s2, e2, s1 == s2, kernel, cutoffs, hyps_mask),
                ),
            )
        )

    # Run child processes.
    for c in children:
        c.start()

    # Wait for all results to arrive.
    matrix = np.zeros((size1, size2))
    for _ in range(nbatch):
        wid, result_chunk = result_queue.get(block=True)
        s1, e1, s2, e2 = block_id[wid]
        matrix[s1 * m1 : e1 * m1, s2 * m2 : e2 * m2] = result_chunk
        # Note that the force/energy block is not symmetric; in this case
        # symm is False.
        if (s1 != s2) and (symm is True):
            matrix[s2 * m2 : e2 * m2, s1 * m1 : e1 * m1] = result_chunk.T

    # Join child processes (clean up zombies).
    for c in children:
        c.join()

    # matrix manipulation
    del result_queue
    del children

    return matrix


def parallel_vector_construction(
    pack_function,
    name,
    x,
    kernel,
    hyps,
    cutoffs,
    hyps_mask,
    block_id,
    nbatch,
    size,
    mult,
    d_1=None,
):
    result_queue = mp.Queue()
    children = []
    for wid in range(nbatch):
        s, e = block_id[wid]
        children.append(
            mp.Process(
                target=queue_wrapper,
                args=(
                    result_queue,
                    wid,
                    pack_function,
                    (name, s, e, x, kernel, hyps, cutoffs, hyps_mask, d_1),
                ),
            )
        )

    # Run child processes.
    for c in children:
        c.start()

    # Wait for all results to arrive.
    vector = np.zeros(size * mult)
    for _ in range(nbatch):
        wid, result_chunk = result_queue.get(block=True)
        s, e = block_id[wid]
        vector[s * mult : e * mult] = result_chunk

    # Join child processes (clean up zombies).
    for c in children:
        c.join()

    return vector


def multiple_array_construction(
    pack_function,
    name,
    x,
    kernel,
    hyps,
    cutoffs,
    hyps_mask,
    block_id,
    nbatch,
    size,
    mult,
    array_sizes,
):
    result_queue = mp.Queue()
    children = []
    for wid in range(nbatch):
        s, e = block_id[wid]
        children.append(
            mp.Process(
                target=queue_wrapper,
                args=(
                    result_queue,
                    wid,
                    pack_function,
                    (name, s, e, x, kernel, hyps, cutoffs, hyps_mask),
                ),
            )
        )

    # Run child processes.
    for c in children:
        c.start()

    # Initialize arrays.
    n_arrays = len(array_sizes)
    arrays = []
    for n in range(n_arrays):
        arrays.append(np.zeros(array_sizes[n]))

    # Wait for all results to arrive.
    for _ in range(nbatch):
        wid, result_chunk = result_queue.get(block=True)
        s, e = block_id[wid]
        for n in range(n_arrays):
            arrays[n][:, s * mult : e * mult] = result_chunk[n]

    # Join child processes (clean up zombies).
    for c in children:
        c.join()

    return arrays


# --------------------------------------------------------------------------
#                            Ky construction
# --------------------------------------------------------------------------


def get_distance_mat_pack(
    hyps: np.ndarray,
    name: str,
    s1: int,
    e1: int,
    s2: int,
    e2: int,
    same: bool,
    kernel,
    cutoffs,
    hyps_mask,
):
    """Compute covariance matrix element between set1 and set2
    :param hyps: list of hyper-parameters
    :param name: name of the gp instance.
    :param same: whether the row and column are the same
    :param kernel: function object of the kernel
    :param cutoffs: The cutoff values used for the atomic environments
    :type cutoffs: list of 2 float numbers
    :param hyps_mask: dictionary used for multi-group hyperparmeters
    :return: covariance matrix
    """

    # initialize matrices
    training_data = _global_training_data[name]
    size1 = e1 - s1
    size2 = e2 - s2
    dist_mat = np.zeros([size1, size2])

    # calculate elements
    args = from_mask_to_args(hyps, cutoffs, hyps_mask)

    for m_index in range(size1):
        x_1 = training_data[m_index + s1]
        if same:
            lowbound = m_index
        else:
            lowbound = 0
        for n_index in range(lowbound, size2):
            x_2 = training_data[n_index + s2]
            kern_curr = kernel(x_1, x_2, *args)
            # store kernel value
            dist_mat[m_index, n_index] = kern_curr
            if same:
                dist_mat[n_index, m_index] = kern_curr

    return dist_mat


def kernel_distance_mat(
    hyps: np.ndarray,
    name: str,
    kernel,
    cutoffs=None,
    hyps_mask=None,
    n_cpus=1,
    n_sample=100,
):
    """parallel version of get_ky_mat
    :param hyps: list of hyper-parameters
    :param name: name of the gp instance.
    :param kernel: function object of the kernel
    :param cutoffs: The cutoff values used for the atomic environments
    :type cutoffs: list of 2 float numbers
    :param hyps_mask: dictionary used for multi-group hyperparmeters
    :return: covariance matrix
    """

    training_data = _global_training_data[name]
    size = len(training_data)

    if n_cpus is None:
        n_cpus = mp.cpu_count()
    if n_cpus == 1:
        k_mat = get_distance_mat_pack(
            hyps, name, 0, size, 0, size, True, kernel, cutoffs, hyps_mask
        )
    else:
        # initialize matrices
        block_id, nbatch = partition_matrix(n_sample, size, n_cpus)
        mult = 1

        k_mat = parallel_matrix_construction(
            get_distance_mat_pack,
            hyps,
            name,
            kernel,
            cutoffs,
            hyps_mask,
            block_id,
            nbatch,
            size,
            size,
            mult,
            mult,
        )

    return k_mat


def get_force_block_pack(
    hyps: np.ndarray,
    name: str,
    s1: int,
    e1: int,
    s2: int,
    e2: int,
    same: bool,
    kernel,
    cutoffs,
    hyps_mask,
):
    """Compute covariance matrix element between set1 and set2
    :param hyps: list of hyper-parameters
    :param name: name of the gp instance.
    :param same: whether the row and column are the same
    :param kernel: function object of the kernel
    :param cutoffs: The cutoff values used for the atomic environments
    :type cutoffs: list of 2 float numbers
    :param hyps_mask: dictionary used for multi-group hyperparmeters

    :return: covariance matrix
    """

    # initialize matrices
    training_data = _global_training_data[name]
    size1 = (e1 - s1) * 3
    size2 = (e2 - s2) * 3
    force_block = np.zeros([size1, size2])

    ds = [1, 2, 3]

    # calculate elements
    args = from_mask_to_args(hyps, cutoffs, hyps_mask)

    for m_index in range(size1):
        x_1 = training_data[int(math.floor(m_index / 3)) + s1]
        d_1 = ds[m_index % 3]
        if same:
            lowbound = m_index
        else:
            lowbound = 0
        for n_index in range(lowbound, size2):
            x_2 = training_data[int(math.floor(n_index / 3)) + s2]
            d_2 = ds[n_index % 3]
            kern_curr = kernel(x_1, x_2, d_1, d_2, *args)
            # store kernel value
            force_block[m_index, n_index] = kern_curr
            if same:
                force_block[n_index, m_index] = kern_curr

    return force_block


def get_energy_block_pack(
    hyps: np.ndarray,
    name: str,
    s1: int,
    e1: int,
    s2: int,
    e2: int,
    same: bool,
    kernel,
    cutoffs,
    hyps_mask,
):

    # initialize matrices
    training_structures = _global_training_structures[name]
    size1 = e1 - s1
    size2 = e2 - s2
    energy_block = np.zeros([size1, size2])

    # calculate elements
    args = from_mask_to_args(hyps, cutoffs, hyps_mask)

    for m_index in range(size1):
        struc_1 = training_structures[m_index + s1]
        if same:
            lowbound = m_index
        else:
            lowbound = 0

        for n_index in range(lowbound, size2):
            struc_2 = training_structures[n_index + s2]

            # Loop over environments in both structures to compute the
            # energy/energy kernel.
            kern_curr = 0
            for environment_1 in struc_1:
                for environment_2 in struc_2:
                    kern_curr += kernel(environment_1, environment_2, *args)

            # Store kernel value.
            energy_block[m_index, n_index] = kern_curr
            if same:
                energy_block[n_index, m_index] = kern_curr

    return energy_block


def get_force_energy_block_pack(
    hyps: np.ndarray,
    name: str,
    s1: int,
    e1: int,
    s2: int,
    e2: int,
    same: bool,
    kernel,
    cutoffs,
    hyps_mask,
):
    # initialize matrices
    training_data = _global_training_data[name]
    training_structures = _global_training_structures[name]
    size1 = (e1 - s1) * 3
    size2 = e2 - s2
    force_energy_block = np.zeros([size1, size2])

    ds = [1, 2, 3]

    # calculate elements
    args = from_mask_to_args(hyps, cutoffs, hyps_mask)

    for m_index in range(size1):
        environment_1 = training_data[int(math.floor(m_index / 3)) + s1]
        d_1 = ds[m_index % 3]

        for n_index in range(size2):
            structure = training_structures[n_index + s2]

            # Loop over environments in the training structure.
            kern_curr = 0
            for environment_2 in structure:
                kern_curr += kernel(environment_1, environment_2, d_1, *args)

            # store kernel value
            force_energy_block[m_index, n_index] = kern_curr

    return force_energy_block


def get_force_block(
    hyps: np.ndarray,
    name: str,
    kernel,
    cutoffs=None,
    hyps_mask=None,
    n_cpus=1,
    n_sample=100,
):
    """parallel version of get_ky_mat
    :param hyps: list of hyper-parameters
    :param name: name of the gp instance.
    :param kernel: function object of the kernel
    :param cutoffs: The cutoff values used for the atomic environments
    :type cutoffs: list of 2 float numbers
    :param hyps_mask: dictionary used for multi-group hyperparmeters

    :return: covariance matrix
    """

    training_data = _global_training_data[name]
    size = len(training_data)
    size3 = 3 * size

    if n_cpus is None:
        n_cpus = mp.cpu_count()
    if n_cpus == 1:
        k_mat = get_force_block_pack(
            hyps, name, 0, size, 0, size, True, kernel, cutoffs, hyps_mask
        )
    else:
        # initialize matrices
        block_id, nbatch = partition_matrix(n_sample, size, n_cpus)
        mult = 3

        k_mat = parallel_matrix_construction(
            get_force_block_pack,
            hyps,
            name,
            kernel,
            cutoffs,
            hyps_mask,
            block_id,
            nbatch,
            size3,
            size3,
            mult,
            mult,
        )

    sigma_n, _, __ = obtain_noise_len(hyps, hyps_mask)
    force_block = k_mat
    force_block += sigma_n ** 2 * np.eye(size3)

    return force_block


def get_energy_block(
    hyps: np.ndarray,
    name: str,
    kernel,
    energy_noise,
    cutoffs=None,
    hyps_mask=None,
    n_cpus=1,
    n_sample=100,
):
    training_structures = _global_training_structures[name]
    size = len(training_structures)

    if n_cpus is None:
        n_cpus = mp.cpu_count()
    if n_cpus == 1:
        k_mat = get_energy_block_pack(
            hyps, name, 0, size, 0, size, True, kernel, cutoffs, hyps_mask
        )
    else:
        block_id, nbatch = partition_matrix(n_sample, size, n_cpus)
        mult = 1

        k_mat = parallel_matrix_construction(
            get_energy_block_pack,
            hyps,
            name,
            kernel,
            cutoffs,
            hyps_mask,
            block_id,
            nbatch,
            size,
            size,
            mult,
            mult,
        )

    energy_block = k_mat
    energy_block += (energy_noise ** 2) * np.eye(size)

    return energy_block


def get_force_energy_block(
    hyps: np.ndarray,
    name: str,
    kernel,
    cutoffs=None,
    hyps_mask=None,
    n_cpus=1,
    n_sample=100,
):
    training_data = _global_training_data[name]
    training_structures = _global_training_structures[name]
    size1 = len(training_data) * 3
    size2 = len(training_structures)
    size3 = len(training_data)

    if n_cpus is None:
        n_cpus = mp.cpu_count()
    if n_cpus == 1:
        force_energy_block = get_force_energy_block_pack(
            hyps, name, 0, size3, 0, size2, True, kernel, cutoffs, hyps_mask
        )
    else:
        # initialize matrices
        block_id, nbatch = partition_force_energy_block(n_sample, size3, size2, n_cpus)
        m1 = 3  # 3 force components per environment
        m2 = 1  # 1 energy per structure

        force_energy_block = parallel_matrix_construction(
            get_force_energy_block_pack,
            hyps,
            name,
            kernel,
            cutoffs,
            hyps_mask,
            block_id,
            nbatch,
            size1,
            size2,
            m1,
            m2,
            symm=False,
        )

    return force_energy_block


def get_Ky_mat(
    hyps: np.ndarray,
    name: str,
    force_kernel: Callable,
    energy_kernel: Callable,
    force_energy_kernel: Callable,
    energy_noise,
    cutoffs=None,
    hyps_mask=None,
    n_cpus=1,
    n_sample=100,
):

    training_data = _global_training_data[name]
    training_structures = _global_training_structures[name]
    size1 = len(training_data) * 3
    size2 = len(training_structures)

    # Initialize Ky.
    ky_mat = np.zeros((size1 + size2, size1 + size2))

    # Assemble the full covariance matrix block-by-block.
    force_block = get_force_block(
        hyps, name, force_kernel, cutoffs, hyps_mask, n_cpus, n_sample
    )

    energy_block = get_energy_block(
        hyps, name, energy_kernel, energy_noise, cutoffs, hyps_mask, n_cpus, n_sample
    )

    force_energy_block = get_force_energy_block(
        hyps, name, force_energy_kernel, cutoffs, hyps_mask, n_cpus, n_sample
    )

    ky_mat[0:size1, 0:size1] = force_block
    ky_mat[size1:, size1:] = energy_block
    ky_mat[0:size1, size1:] = force_energy_block
    ky_mat[size1:, 0:size1] = force_energy_block.transpose()

    return ky_mat


# --------------------------------------------------------------------------
#                              Ky updates
# --------------------------------------------------------------------------


def update_force_block(
    ky_mat_old: np.ndarray,
    n_envs_prev: int,
    hyps: np.ndarray,
    name: str,
    kernel,
    cutoffs=None,
    hyps_mask=None,
    n_cpus=1,
    n_sample=100,
):

    old_size = n_envs_prev
    old_size3 = n_envs_prev * 3
    mult = 3
    size = len(_global_training_data[name])
    size3 = 3 * size

    if n_cpus is None:
        n_cpus = mp.cpu_count()

    # serial version
    if n_cpus == 1:
        force_block = np.zeros((size3, size3))
        new_mat = get_force_block_pack(
            hyps, name, 0, size, old_size, size, False, kernel, cutoffs, hyps_mask
        )
        force_block[:, old_size3:] = new_mat
        force_block[old_size3:, :] = new_mat.transpose()

    # parallel version
    else:
        block_id, nbatch = partition_update(n_sample, size, old_size, n_cpus)

        force_block = parallel_matrix_construction(
            get_force_block_pack,
            hyps,
            name,
            kernel,
            cutoffs,
            hyps_mask,
            block_id,
            nbatch,
            size3,
            size3,
            mult,
            mult,
        )

    # insert previous covariance matrix
    force_block[:old_size3, :old_size3] = ky_mat_old[:old_size3, :old_size3]

    # add the noise parameter
    sigma_n, _, _ = obtain_noise_len(hyps, hyps_mask)
    force_block[old_size3:, old_size3:] += sigma_n ** 2 * np.eye(size3 - old_size3)

    return force_block


def update_energy_block(
    ky_mat_old: np.ndarray,
    n_envs_prev: int,
    hyps: np.ndarray,
    name: str,
    kernel,
    energy_noise: float,
    cutoffs=None,
    hyps_mask=None,
    n_cpus=1,
    n_sample=100,
):

    old_size = ky_mat_old.shape[0] - 3 * n_envs_prev
    mult = 1
    size = len(_global_training_structures[name])

    if n_cpus is None:
        n_cpus = mp.cpu_count()

    # serial version
    if n_cpus == 1:
        energy_block = np.zeros((size, size))
        new_mat = get_energy_block_pack(
            hyps, name, 0, size, old_size, size, False, kernel, cutoffs, hyps_mask
        )
        energy_block[:, old_size:] = new_mat
        energy_block[old_size:, :] = new_mat.transpose()

    # parallel version
    else:
        block_id, nbatch = partition_update(n_sample, size, old_size, n_cpus)

        energy_block = parallel_matrix_construction(
            get_energy_block_pack,
            hyps,
            name,
            kernel,
            cutoffs,
            hyps_mask,
            block_id,
            nbatch,
            size,
            size,
            mult,
            mult,
        )

    # insert previous covariance matrix (if it has nonzero size)
    if old_size > 0:
        energy_block[:old_size, :old_size] = ky_mat_old[-old_size:, -old_size:]

    # add the noise parameter
    energy_block[old_size:, old_size:] += (energy_noise ** 2) * np.eye(size - old_size)

    return energy_block


def update_force_energy_block(
    ky_mat_old: np.ndarray,
    n_envs_prev: int,
    hyps: np.ndarray,
    name: str,
    kernel,
    energy_noise: float,
    cutoffs=None,
    hyps_mask=None,
    n_cpus=1,
    n_sample=100,
):

    n_strucs_prev = ky_mat_old.shape[0] - 3 * n_envs_prev
    n_envs = len(_global_training_data[name])
    n_strucs = len(_global_training_structures[name])
    force_energy_block = np.zeros((n_envs * 3, n_strucs))

    if n_cpus is None:
        n_cpus = mp.cpu_count()

    # serial version
    if n_cpus == 1:
        mat_1 = get_force_energy_block_pack(
            hyps,
            name,
            0,
            n_envs,
            n_strucs_prev,
            n_strucs,
            False,
            kernel,
            cutoffs,
            hyps_mask,
        )

        mat_2 = get_force_energy_block_pack(
            hyps,
            name,
            n_envs_prev,
            n_envs,
            0,
            n_strucs_prev,
            False,
            kernel,
            cutoffs,
            hyps_mask,
        )

    # parallel version
    else:
        force_energy_block = np.zeros((n_envs * 3, n_strucs))

        block_id_1, nbatch_1 = partition_matrix_custom(
            n_sample, 0, n_envs, n_strucs_prev, n_strucs, n_cpus
        )
        block_id_2, nbatch_2 = partition_matrix_custom(
            n_sample, n_envs_prev, n_envs, 0, n_strucs_prev, n_cpus
        )

        size1 = n_envs * 3
        size2 = n_strucs
        m1 = 3
        m2 = 1

        mat_1 = parallel_matrix_construction(
            get_force_energy_block_pack,
            hyps,
            name,
            kernel,
            cutoffs,
            hyps_mask,
            block_id_1,
            nbatch_1,
            size1,
            size2,
            m1,
            m2,
            False,
        )
        mat_2 = parallel_matrix_construction(
            get_force_energy_block_pack,
            hyps,
            name,
            kernel,
            cutoffs,
            hyps_mask,
            block_id_2,
            nbatch_2,
            size1,
            size2,
            m1,
            m2,
            False,
        )

        # reduce the size of the matrices
        mat_1 = mat_1[: n_envs * 3, n_strucs_prev:n_strucs]
        mat_2 = mat_2[n_envs_prev * 3 : n_envs * 3, :n_strucs_prev]

    force_energy_block[: n_envs * 3, n_strucs_prev:n_strucs] = mat_1
    force_energy_block[n_envs_prev * 3 : n_envs * 3, :n_strucs_prev] = mat_2

    # insert previous covariance matrix
    force_energy_block[: n_envs_prev * 3, :n_strucs_prev] = ky_mat_old[
        : n_envs_prev * 3, n_envs_prev * 3 : n_envs_prev * 3 + n_strucs_prev
    ]

    return force_energy_block


def get_ky_mat_update(
    ky_mat_old: np.ndarray,
    n_envs_prev: int,
    hyps: np.ndarray,
    name: str,
    force_kernel: Callable,
    energy_kernel: Callable,
    force_energy_kernel: Callable,
    energy_noise: float,
    cutoffs=None,
    hyps_mask=None,
    n_cpus=1,
    n_sample=100,
):

    n_envs = len(_global_training_data[name])
    n_strucs = len(_global_training_structures[name])
    size1 = n_envs * 3
    size2 = n_strucs
    ky_mat = np.zeros((size1 + size2, size1 + size2))

    force_block = update_force_block(
        ky_mat_old,
        n_envs_prev,
        hyps,
        name,
        force_kernel,
        cutoffs,
        hyps_mask,
        n_cpus,
        n_sample,
    )

    energy_block = update_energy_block(
        ky_mat_old,
        n_envs_prev,
        hyps,
        name,
        energy_kernel,
        energy_noise,
        cutoffs,
        hyps_mask,
        n_cpus,
        n_sample,
    )

    force_energy_block = update_force_energy_block(
        ky_mat_old,
        n_envs_prev,
        hyps,
        name,
        force_energy_kernel,
        energy_noise,
        cutoffs,
        hyps_mask,
        n_cpus,
        n_sample,
    )

    ky_mat[0:size1, 0:size1] = force_block
    ky_mat[size1:, size1:] = energy_block
    ky_mat[0:size1, size1:] = force_energy_block
    ky_mat[size1:, 0:size1] = force_energy_block.transpose()

    return ky_mat


# --------------------------------------------------------------------------
#                            Kernel vectors
# --------------------------------------------------------------------------


def energy_energy_vector_unit(
    name, s, e, x, kernel, hyps, cutoffs=None, hyps_mask=None, d_1=None
):
    """
    Gets part of the energy/energy vector.
    """

    training_structures = _global_training_structures[name]

    size = e - s
    energy_energy_unit = np.zeros(
        size,
    )

    args = from_mask_to_args(hyps, cutoffs, hyps_mask)

    for m_index in range(size):
        structure = training_structures[m_index + s]
        kern_curr = 0
        for environment in structure:
            kern_curr += kernel(x, environment, *args)

        energy_energy_unit[m_index] = kern_curr

    return energy_energy_unit


def energy_force_vector_unit(
    name, s, e, x, kernel, hyps, cutoffs=None, hyps_mask=None, d_1=None
):
    """
    Gets part of the energy/force vector.
    """

    training_data = _global_training_data[name]

    ds = [1, 2, 3]
    size = (e - s) * 3
    k_v = np.zeros(
        size,
    )

    args = from_mask_to_args(hyps, cutoffs, hyps_mask)

    for m_index in range(size):
        x_2 = training_data[int(math.floor(m_index / 3)) + s]
        d_2 = ds[m_index % 3]
        k_v[m_index] = kernel(x_2, x, d_2, *args)

    return k_v


def force_energy_vector_unit(name, s, e, x, kernel, hyps, cutoffs, hyps_mask, d_1):
    """
    Gets part of the force/energy vector.
    """

    training_structures = _global_training_structures[name]

    size = e - s
    args = from_mask_to_args(hyps, cutoffs, hyps_mask)
    force_energy_unit = np.zeros(
        size,
    )

    for m_index in range(size):
        training_structure = training_structures[m_index + s]
        kern_curr = 0
        for environment in training_structure:
            kern_curr += kernel(x, environment, d_1, *args)

        force_energy_unit[m_index] = kern_curr

    return force_energy_unit


def force_force_vector_unit(name, s, e, x, kernel, hyps, cutoffs, hyps_mask, d_1):
    """
    Gets part of the force/force vector.
    """

    training_data = _global_training_data[name]

    size = e - s
    ds = [1, 2, 3]

    args = from_mask_to_args(hyps, cutoffs, hyps_mask)

    k_v = np.zeros(size * 3)

    for m_index in range(size):
        x_2 = training_data[m_index + s]
        for d_2 in ds:
            k_v[m_index * 3 + d_2 - 1] = kernel(x, x_2, d_1, d_2, *args)

    return k_v


def efs_force_vector_unit(name, s, e, x, efs_force_kernel, hyps, cutoffs, hyps_mask):

    training_data = _global_training_data[name]

    size = e - s
    k_ef = np.zeros((1, size * 3))
    k_ff = np.zeros((3, size * 3))
    k_sf = np.zeros((6, size * 3))

    args = from_mask_to_args(hyps, cutoffs, hyps_mask)

    for m_index in range(size):
        x_2 = training_data[m_index + s]
        ef, ff, sf = efs_force_kernel(x, x_2, *args)

        ind1 = m_index * 3
        ind2 = (m_index + 1) * 3

        k_ef[:, ind1:ind2] = ef
        k_ff[:, ind1:ind2] = ff
        k_sf[:, ind1:ind2] = sf

    return k_ef, k_ff, k_sf


def efs_energy_vector_unit(name, s, e, x, efs_energy_kernel, hyps, cutoffs, hyps_mask):

    training_structures = _global_training_structures[name]

    size = e - s
    args = from_mask_to_args(hyps, cutoffs, hyps_mask)

    k_ee = np.zeros((1, size))
    k_fe = np.zeros((3, size))
    k_se = np.zeros((6, size))

    for m_index in range(size):
        training_structure = training_structures[m_index + s]

        ee_curr = 0
        fe_curr = np.zeros(3)
        se_curr = np.zeros(6)

        for environment in training_structure:
            ee, fe, se = efs_energy_kernel(x, environment, *args)
            ee_curr += ee
            fe_curr += fe
            se_curr += se

        k_ee[:, m_index] = ee_curr
        k_fe[:, m_index] = fe_curr
        k_se[:, m_index] = se_curr

    return k_ee, k_fe, k_se


def energy_energy_vector(
    name, kernel, x, hyps, cutoffs=None, hyps_mask=None, n_cpus=1, n_sample=100
):
    """
    Get a vector of covariances between the local energy of a test environment
    and the total energy labels in the training set.
    """

    size = len(_global_training_structures[name])

    if n_cpus is None:
        n_cpus = mp.cpu_count()
    if n_cpus == 1:
        return energy_energy_vector_unit(
            name, 0, size, x, kernel, hyps, cutoffs, hyps_mask
        )

    block_id, nbatch = partition_vector(n_sample, size, n_cpus)
    pack_function = energy_energy_vector_unit
    mult = 1

    force_energy_vector = parallel_vector_construction(
        pack_function,
        name,
        x,
        kernel,
        hyps,
        cutoffs,
        hyps_mask,
        block_id,
        nbatch,
        size,
        mult,
    )

    return force_energy_vector


def energy_force_vector(
    name, kernel, x, hyps, cutoffs=None, hyps_mask=None, n_cpus=1, n_sample=100
):
    """
    Get the vector of covariances between the local energy of a test
    environment and the force labels in the training set.
    """

    training_data = _global_training_data[name]
    size = len(training_data)

    if n_cpus is None:
        n_cpus = mp.cpu_count()
    if n_cpus == 1:
        return energy_force_vector_unit(
            name, 0, size, x, kernel, hyps, cutoffs, hyps_mask
        )

    block_id, nbatch = partition_vector(n_sample, size, n_cpus)
    pack_function = energy_force_vector_unit
    mult = 3

    k12_v = parallel_vector_construction(
        pack_function,
        name,
        x,
        kernel,
        hyps,
        cutoffs,
        hyps_mask,
        block_id,
        nbatch,
        size,
        mult,
    )

    return k12_v


def force_energy_vector(
    name, kernel, x, d_1, hyps, cutoffs=None, hyps_mask=None, n_cpus=1, n_sample=100
):
    """
    Get a vector of covariances between a force component of a test environment
    and the total energy labels in the training set.
    """

    size = len(_global_training_structures[name])

    if n_cpus is None:
        n_cpus = mp.cpu_count()
    if n_cpus == 1:
        return force_energy_vector_unit(
            name, 0, size, x, kernel, hyps, cutoffs, hyps_mask, d_1
        )

    block_id, nbatch = partition_vector(n_sample, size, n_cpus)
    pack_function = force_energy_vector_unit
    mult = 1

    force_energy_vector = parallel_vector_construction(
        pack_function,
        name,
        x,
        kernel,
        hyps,
        cutoffs,
        hyps_mask,
        block_id,
        nbatch,
        size,
        mult,
        d_1,
    )

    return force_energy_vector


def force_force_vector(
    name, kernel, x, d_1, hyps, cutoffs=None, hyps_mask=None, n_cpus=1, n_sample=100
):
    """
    Get a vector of covariances between a force component of a test environment
    and the force labels in the training set.
    """

    size = len(_global_training_data[name])

    if n_cpus is None:
        n_cpus = mp.cpu_count()
    if n_cpus == 1:
        return force_force_vector_unit(
            name, 0, size, x, kernel, hyps, cutoffs, hyps_mask, d_1
        )

    block_id, nbatch = partition_vector(n_sample, size, n_cpus)
    pack_function = force_force_vector_unit
    mult = 3

    k12_v = parallel_vector_construction(
        pack_function,
        name,
        x,
        kernel,
        hyps,
        cutoffs,
        hyps_mask,
        block_id,
        nbatch,
        size,
        mult,
        d_1,
    )

    return k12_v


def efs_force_vector(
    name,
    efs_force_kernel,
    x,
    hyps,
    cutoffs=None,
    hyps_mask=None,
    n_cpus=1,
    n_sample=100,
):
    """
    Returns covariances between the local eneregy, force components, and
    partial stresses of a test environment and the force labels in the
    training set.
    """

    size = len(_global_training_data[name])

    if n_cpus is None:
        n_cpus = mp.cpu_count()

    # Perform serial calculation if n_cpus = 1.
    if n_cpus == 1:
        return efs_force_vector_unit(
            name, 0, size, x, efs_force_kernel, hyps, cutoffs, hyps_mask
        )

    # Otherwise, perform parallel calculation.
    block_id, nbatch = partition_vector(n_sample, size, n_cpus)
    pack_function = efs_force_vector_unit
    mult = 3
    n_comps = size * mult
    array_sizes = [(1, n_comps), (3, n_comps), (6, n_comps)]

    efs_arrays = multiple_array_construction(
        pack_function,
        name,
        x,
        efs_force_kernel,
        hyps,
        cutoffs,
        hyps_mask,
        block_id,
        nbatch,
        size,
        mult,
        array_sizes,
    )

    return efs_arrays


def efs_energy_vector(
    name,
    efs_energy_kernel,
    x,
    hyps,
    cutoffs=None,
    hyps_mask=None,
    n_cpus=1,
    n_sample=100,
):
    """
    Returns covariances between the local eneregy, force components, and
    partial stresses of a test environment and the total energy labels in the
    training set.
    """

    size = len(_global_training_structures[name])

    if n_cpus is None:
        n_cpus = mp.cpu_count()

    # Perform serial calculation if n_cpus = 1.
    if n_cpus == 1:
        return efs_energy_vector_unit(
            name, 0, size, x, efs_energy_kernel, hyps, cutoffs, hyps_mask
        )

    # Otherwise, perform parallel calculation.
    block_id, nbatch = partition_vector(n_sample, size, n_cpus)
    pack_function = efs_energy_vector_unit
    mult = 1
    array_sizes = [(1, size), (3, size), (6, size)]

    efs_arrays = multiple_array_construction(
        pack_function,
        name,
        x,
        efs_energy_kernel,
        hyps,
        cutoffs,
        hyps_mask,
        block_id,
        nbatch,
        size,
        mult,
        array_sizes,
    )

    return efs_arrays


def get_kernel_vector(
    name,
    force_force_kernel: Callable,
    force_energy_kernel: Callable,
    x,
    d_1,
    hyps,
    cutoffs=None,
    hyps_mask=None,
    n_cpus=1,
    n_sample=100,
):

    size1 = len(_global_training_data[name])
    size2 = len(_global_training_structures[name])
    kernel_vector = np.zeros(size1 * 3 + size2)

    force_vector = force_force_vector(
        name, force_force_kernel, x, d_1, hyps, cutoffs, hyps_mask, n_cpus, n_sample
    )
    energy_vector = force_energy_vector(
        name, force_energy_kernel, x, d_1, hyps, cutoffs, hyps_mask, n_cpus, n_sample
    )

    kernel_vector[0 : size1 * 3] = force_vector
    kernel_vector[size1 * 3 :] = energy_vector

    return kernel_vector


def en_kern_vec(
    name,
    energy_force_kernel,
    energy_energy_kernel,
    x,
    hyps,
    cutoffs=None,
    hyps_mask=None,
    n_cpus=1,
    n_sample=100,
):

    size1 = len(_global_training_data[name])
    size2 = len(_global_training_structures[name])
    kernel_vector = np.zeros(size1 * 3 + size2)

    force_vector = energy_force_vector(
        name, energy_force_kernel, x, hyps, cutoffs, hyps_mask, n_cpus, n_sample
    )
    energy_vector = energy_energy_vector(
        name, energy_energy_kernel, x, hyps, cutoffs, hyps_mask, n_cpus, n_sample
    )

    kernel_vector[0 : size1 * 3] = force_vector
    kernel_vector[size1 * 3 :] = energy_vector

    return kernel_vector


def efs_kern_vec(
    name,
    efs_force_kernel,
    efs_energy_kernel,
    x,
    hyps,
    cutoffs=None,
    hyps_mask=None,
    n_cpus=1,
    n_sample=100,
):

    size1 = len(_global_training_data[name])
    size2 = len(_global_training_structures[name])
    tot_size = size1 * 3 + size2

    # Initialize arrays.
    energy_vector = np.zeros(tot_size)
    force_array = np.zeros((3, tot_size))
    stress_array = np.zeros((6, tot_size))

    # Compute force and energy arrays.
    force_arrays = efs_force_vector(
        name, efs_force_kernel, x, hyps, cutoffs, hyps_mask, n_cpus, n_sample
    )

    energy_arrays = efs_energy_vector(
        name, efs_energy_kernel, x, hyps, cutoffs, hyps_mask, n_cpus, n_sample
    )

    # Populate arrays.
    energy_vector[0 : size1 * 3] = force_arrays[0]
    energy_vector[size1 * 3 :] = energy_arrays[0]
    force_array[:, 0 : size1 * 3] = force_arrays[1]
    force_array[:, size1 * 3 :] = energy_arrays[1]
    stress_array[:, 0 : size1 * 3] = force_arrays[2]
    stress_array[:, size1 * 3 :] = energy_arrays[2]

    return energy_vector, force_array, stress_array


# --------------------------------------------------------------------------
#                      Ky hyperparameter gradient
# --------------------------------------------------------------------------


def get_ky_and_hyp_pack(
    name,
    s1,
    e1,
    s2,
    e2,
    same: bool,
    hyps: np.ndarray,
    kernel_grad,
    cutoffs=None,
    hyps_mask=None,
):
    """
    computes a block of ky matrix and its derivative to hyper-parameter
    If the cpu set up is None, it uses as much as posible cpus

    :param hyps: list of hyper-parameters
    :param name: name of the gp instance.
    :param kernel_grad: function object of the kernel gradient
    :param cutoffs: The cutoff values used for the atomic environments
    :type cutoffs: list of 2 float numbers
    :param hyps_mask: dictionary used for multi-group hyperparmeters

    :return: hyp_mat, ky_mat
    """

    # assume sigma_n is the final hyperparameter
    _, non_noise_hyps, _ = obtain_noise_len(hyps, hyps_mask)

    # initialize matrices
    size1 = (e1 - s1) * 3
    size2 = (e2 - s2) * 3
    k_mat = np.zeros([size1, size2])
    hyp_mat = np.zeros([non_noise_hyps, size1, size2])

    args = from_mask_to_args(hyps, cutoffs, hyps_mask)

    ds = [1, 2, 3]

    training_data = _global_training_data[name]
    # calculate elements
    for m_index in range(size1):
        x_1 = training_data[int(math.floor(m_index / 3)) + s1]
        d_1 = ds[m_index % 3]

        if same:
            lowbound = m_index
        else:
            lowbound = 0
        for n_index in range(lowbound, size2):
            x_2 = training_data[int(math.floor(n_index / 3)) + s2]
            d_2 = ds[n_index % 3]

            # calculate kernel and gradient
            cov = kernel_grad(x_1, x_2, d_1, d_2, *args)

            # store kernel value
            k_mat[m_index, n_index] = cov[0]
            grad = from_grad_to_mask(cov[1], hyps_mask)

            hyp_mat[:, m_index, n_index] = grad
            if same:
                k_mat[n_index, m_index] = cov[0]
                hyp_mat[:, n_index, m_index] = grad

    return hyp_mat, k_mat


def get_ky_and_hyp(
    hyps: np.ndarray,
    name,
    kernel_grad,
    cutoffs=None,
    hyps_mask=None,
    n_cpus=1,
    n_sample=100,
):
    """
    parallel version of get_ky_and_hyp

    :param hyps: list of hyper-parameters
    :param name: name of the gp instance.
    :param kernel_grad: function object of the kernel gradient
    :param cutoffs: The cutoff values used for the atomic environments
    :type cutoffs: list of 2 float numbers
    :param hyps_mask: dictionary used for multi-group hyperparmeters
    :param n_cpus: number of cpus to use.
    :param n_sample: the size of block for matrix to compute

    :return: hyp_mat, ky_mat
    """

    training_data = _global_training_data[name]
    size = len(training_data)
    size3 = size * 3

    sigma_n, non_noise_hyps, train_noise = obtain_noise_len(hyps, hyps_mask)

    if n_cpus is None:
        n_cpus = mp.cpu_count()
    if n_cpus == 1:
        hyp_mat0, k_mat = get_ky_and_hyp_pack(
            name, 0, size, 0, size, True, hyps, kernel_grad, cutoffs, hyps_mask
        )
    else:

        block_id, nbatch = partition_matrix(n_sample, size, n_cpus)

        result_queue = mp.Queue()
        children = []
        for wid in range(nbatch):
            s1, e1, s2, e2 = block_id[wid]
            children.append(
                mp.Process(
                    target=queue_wrapper,
                    args=(
                        result_queue,
                        wid,
                        get_ky_and_hyp_pack,
                        (
                            name,
                            s1,
                            e1,
                            s2,
                            e2,
                            s1 == s2,
                            hyps,
                            kernel_grad,
                            cutoffs,
                            hyps_mask,
                        ),
                    ),
                )
            )

        # Run child processes.
        for c in children:
            c.start()

        k_mat = np.zeros([size3, size3])
        hyp_mat0 = np.zeros([non_noise_hyps, size3, size3])
        for _ in range(nbatch):
            wid, result_chunk = result_queue.get(block=True)
            s1, e1, s2, e2 = block_id[wid]
            h_mat_block, k_mat_block = result_chunk
            k_mat[s1 * 3 : e1 * 3, s2 * 3 : e2 * 3] = k_mat_block
            hyp_mat0[:, s1 * 3 : e1 * 3, s2 * 3 : e2 * 3] = h_mat_block
            if s1 != s2:
                k_mat[s2 * 3 : e2 * 3, s1 * 3 : e1 * 3] = k_mat_block.T
                for idx in range(hyp_mat0.shape[0]):
                    hyp_mat0[idx, s2 * 3 : e2 * 3, s1 * 3 : e1 * 3] = h_mat_block[idx].T

        # Join child processes (clean up zombies).
        for c in children:
            c.join()

    # add gradient of noise variance
    if train_noise:
        hyp_mat = np.zeros(
            [hyp_mat0.shape[0] + 1, hyp_mat0.shape[1], hyp_mat0.shape[2]]
        )
        hyp_mat[:-1, :, :] = hyp_mat0
        hyp_mat[-1, :, :] = np.eye(size3) * 2 * sigma_n
    else:
        hyp_mat = hyp_mat0

    # matrix manipulation
    ky_mat = k_mat + sigma_n ** 2 * np.eye(size3)

    return hyp_mat, ky_mat


# --------------------------------------------------------------------------
#                        Computing the likelihood
# --------------------------------------------------------------------------


def get_like_from_mats(ky_mat, l_mat, alpha, name):
    """compute the likelihood from the covariance matrix

    :param ky_mat: the covariance matrix

    :return: float, likelihood
    """
    # catch linear algebra errors
    force_labels = _global_training_labels[name]
    energy_labels = _global_energy_labels[name]
    labels = np.concatenate((force_labels, energy_labels))

    # calculate likelihood
    like = (
        -0.5 * np.matmul(labels, alpha)
        - np.sum(np.log(np.diagonal(l_mat)))
        - math.log(2 * np.pi) * ky_mat.shape[1] / 2
    )

    return like


def get_neg_like(
    hyps: np.ndarray,
    name: str,
    force_kernel,
    logger_name=None,
    cutoffs=None,
    hyps_mask=None,
    n_cpus=1,
    n_sample=100,
):
    """compute the log likelihood and its gradients
    :param hyps: list of hyper-parameters
    :type hyps: np.ndarray
    :param name: name of the gp instance.
    :param kernel_grad: function object of the kernel gradient
    :param output: Output object for dumping every hyper-parameter
                   sets computed
    :type output: logger_name
    :param cutoffs: The cutoff values used for the atomic environments
    :type cutoffs: list of 2 float numbers
    :param hyps_mask: dictionary used for multi-group hyperparmeters
    :param n_cpus: number of cpus to use.
    :param n_sample: the size of block for matrix to compute
    :return: float, np.array
    """

    time0 = time.time()

    ky_mat = get_force_block(
        hyps,
        name,
        force_kernel,
        cutoffs=cutoffs,
        hyps_mask=hyps_mask,
        n_cpus=n_cpus,
        n_sample=n_sample,
    )

    logger = logging.getLogger(logger_name)
    logger.debug(f"{name} get_ky_and_hyp {time.time()-time0}")

    time0 = time.time()

    # catch linear algebra errors
    try:
        ky_mat_inv = np.linalg.inv(ky_mat)
        l_mat = np.linalg.cholesky(ky_mat)
    except np.linalg.LinAlgError:
        number_of_hyps = len(hyps)
        return -1e8, np.zeros(number_of_hyps)

    labels = _global_training_labels[name]

    alpha = np.matmul(ky_mat_inv, labels)

    like = get_like_from_mats(ky_mat,l_mat,alpha,name)

    logger.debug(f"get_like_from_mats {time.time()-time0}")

    logger.debug(f"{name} Hyperparameters: {list(hyps)}")
    logger.debug(f"{name} Likelihood: {like}")

    return -like

def get_neg_like_grad(
    hyps: np.ndarray,
    name: str,
    kernel_grad,
    logger_name: str = None,
    cutoffs=None,
    hyps_mask=None,
    n_cpus=1,
    n_sample=100,
):
    """compute the log likelihood and its gradients

    :param hyps: list of hyper-parameters
    :type hyps: np.ndarray
    :param name: name of the gp instance.
    :param kernel_grad: function object of the kernel gradient
    :param logger_name: name of logger object for dumping every hyper-parameter
                   sets computed
    :type logger_name: str
    :param cutoffs: The cutoff values used for the atomic environments
    :type cutoffs: list of 2 float numbers
    :param hyps_mask: dictionary used for multi-group hyperparmeters
    :param n_cpus: number of cpus to use.
    :param n_sample: the size of block for matrix to compute

    :return: float, np.array
    """

    time0 = time.time()

    hyp_mat, ky_mat = get_ky_and_hyp(
        hyps,
        name,
        kernel_grad,
        cutoffs=cutoffs,
        hyps_mask=hyps_mask,
        n_cpus=n_cpus,
        n_sample=n_sample,
    )

    logger = logging.getLogger(logger_name)
    logger.debug(f"get_ky_and_hyp {time.time()-time0}")

    time0 = time.time()

    like, like_grad = get_like_grad_from_mats(ky_mat, hyp_mat, name)

    logger.debug(f"get_like_grad_from_mats {time.time()-time0}")

    logger.debug("")
    logger.info(f"Hyperparameters: {list(hyps)}")
    logger.info(f"Likelihood: {like}")
    logger.info(f"Likelihood Gradient: {list(like_grad)}")

    return -like, -like_grad


def get_like_grad_from_mats(ky_mat, hyp_mat, name):
    """compute the gradient of likelihood to hyper-parameters
    from covariance matrix and its gradient

    :param ky_mat: covariance matrix
    :type ky_mat: np.array
    :param hyp_mat: dky/d(hyper parameter) matrix
    :type hyp_mat: np.array
    :param name: name of the gp instance.

    :return: float, list. the likelihood and its gradients
    """

    number_of_hyps = hyp_mat.shape[0]

    # catch linear algebra errors
    try:
        ky_mat_inv = np.linalg.inv(ky_mat)
        l_mat = np.linalg.cholesky(ky_mat)
    except np.linalg.LinAlgError:
        return -1e8, np.zeros(number_of_hyps)

    labels = _global_training_labels[name]

    alpha = np.matmul(ky_mat_inv, labels)
    alpha_mat = np.matmul(alpha.reshape(-1, 1), alpha.reshape(1, -1))
    like_mat = alpha_mat - ky_mat_inv

    # calculate likelihood
    like = (
        -0.5 * np.matmul(labels, alpha)
        - np.sum(np.log(np.diagonal(l_mat)))
        - math.log(2 * np.pi) * ky_mat.shape[1] / 2
    )

    # calculate likelihood gradient
    like_grad = np.zeros(number_of_hyps)
    for n in range(number_of_hyps):
        like_grad[n] = 0.5 * np.trace(np.matmul(like_mat, hyp_mat[n, :, :]))

    return like, like_grad



"""
Utility functions for various tasks.
"""
from warnings import warn
from json import JSONEncoder
from typing import List
from math import inf

import numpy as np


def is_std_in_bound(
    std_tolerance: float,
    noise: float,
    structure: "flare.struc.Structure",
    max_atoms_added: int = inf,
) -> (bool, List[int]):
    """
    Given an uncertainty tolerance and a structure decorated with atoms,
    species, and associated uncertainties, return those which are above a
    given threshold, agnostic to species.

    If std_tolerance is negative, then the threshold used is the absolute
    value of std_tolerance.

    If std_tolerance is positive, then the threshold used is
    std_tolerance * noise.

    If std_tolerance is 0, then do not check.

    :param std_tolerance: If positive, multiply by noise to get cutoff. If
        negative, use absolute value of std_tolerance as cutoff.
    :param noise: Noise variance parameter
    :param structure: Input structure
    :type structure: FLARE Structure
    :param max_atoms_added: Maximum # of atoms to add
    :return: (True,[-1]) if no atoms are above cutoff, (False,[...]) of the
            top `max_atoms_added` uncertainties
    """
    # set uncertainty threshold
    if std_tolerance == 0:
        return True, [-1]
    elif std_tolerance > 0:
        threshold = std_tolerance * np.abs(noise)
    else:
        threshold = np.abs(std_tolerance)

    # sort max stds
    nat = len(structure)
    max_stds = np.zeros((nat))
    for atom, std in enumerate(structure.stds):
        max_stds[atom] = np.max(std)
    stds_sorted = np.argsort(max_stds)
    target_atoms = list(stds_sorted[-max_atoms_added:])

    # if above threshold, return atom
    if max_stds[stds_sorted[-1]] > threshold:
        return False, target_atoms
    else:
        return True, [-1]


def is_std_in_bound_per_species(
    rel_std_tolerance: float,
    abs_std_tolerance: float,
    noise: float,
    structure: "flare.struc.Structure",
    max_atoms_added: int = inf,
    max_by_species: dict = {},
) -> (bool, List[int]):
    """
    Checks the stds of GP prediction assigned to the structure, returns a
    list of atoms which either meet an absolute threshold or a relative
    threshold defined by rel_std_tolerance * noise. Can limit the
    total number of target atoms via max_atoms_added, and limit per species
    by max_by_species.

    The max_atoms_added argument will 'overrule' the
    max by species; e.g. if max_atoms_added is 2 and max_by_species is {"H":3},
    then at most two atoms will be added.

    :param rel_std_tolerance: Multiplied by noise to get a lower
        bound for the uncertainty threshold defined relative to the model.
    :param abs_std_tolerance: Used as an absolute lower bound for the
        uncertainty threshold.
    :param noise: Noise hyperparameter for model, used to define relative
        uncertainty cutoff.
    :param structure: FLARE structure decorated with
        uncertainties in structure.stds.
    :param max_atoms_added: Maximum number of atoms to return from structure.
    :param max_by_species: Dictionary describing maximum number of atoms to
        return by species (e.g. {'H':1,'He':2} will return at most 1 H and 2 He
        atoms.)
    :return: Bool indicating if any atoms exceeded the uncertainty
        threshold, and a list of indices of atoms which did, sorted by their
        uncertainty.
    """

    # Always returns true; use this when you want to test model performance
    # without updating the training set.
    if rel_std_tolerance == 0 and abs_std_tolerance == 0:
        return True, [-1]

    # set uncertainty threshold based on if only one or the other is passed in,
    # and use the lower of the two.

    if rel_std_tolerance is None or rel_std_tolerance == 0:
        threshold = abs_std_tolerance
    elif abs_std_tolerance is None or abs_std_tolerance == 0:
        threshold = rel_std_tolerance * np.abs(noise)
    else:
        threshold = min(rel_std_tolerance * np.abs(noise), abs_std_tolerance)

    # Determine if any std component will trigger the threshold
    # before looking through individual species.
    max_std_components = [np.nanmax(std) for std in structure.stds]
    if np.nanmax(max_std_components) < threshold:
        return True, [-1]

    target_atoms = []

    # Sort from greatest to smallest max. std component
    std_arg_sorted = np.flip(np.argsort(max_std_components))

    present_species = {spec: 0 for spec in set(structure.species_labels)}

    # Loop through atoms and add until cutoffs are met.
    for i in std_arg_sorted:

        # If max atoms added reached or stds of atoms considered are now below
        # threshold, conclude
        if len(target_atoms) == max_atoms_added or (
            max_std_components[i] < threshold and max_std_components[i] != np.nan
        ):
            break

        if np.isnan(max_std_components[i]):
            continue

        # Only add up to species allowance, if it exists
        cur_spec = structure.species_labels[i]
        if present_species[cur_spec] < max_by_species.get(cur_spec, inf):
            target_atoms.append(i)
            present_species[cur_spec] += 1

    # Check in case that nothing was added, e.g. due to species limitations
    if len(target_atoms):
        return False, target_atoms

    return True, [-1]


def is_force_in_bound_per_species(
    abs_force_tolerance: float,
    predicted_forces: "ndarray",
    label_forces: "ndarray",
    structure,
    max_atoms_added: int = inf,
    max_by_species: dict = {},
    max_force_error: float = inf,
) -> (bool, List[int]):
    """
    Checks the forces of GP prediction assigned to the structure against a
    DFT calculation, and return a list of atoms which meet an absolute
    threshold abs_force_tolerance.

    Can limit the total number of target atoms via max_atoms_added, and limit
    per species by max_by_species.

    The max_atoms_added argument will 'overrule' the
    max by species; e.g. if max_atoms_added is 2 and max_by_species is {"H":3},
    then at most two atoms total will be added.

    Because adding atoms which are in configurations which are far outside
    of the potential energy surface may not always be
    desirable, a maximum force error can be passed in; atoms with

    :param abs_force_tolerance: If error exceeds this value, then return
        atom index
    :param predicted_forces: Force predictions made by GP model
    :param label_forces: "True" forces computed by DFT
    :param structure: FLARE Structure
    :param max_atoms_added: Maximum atoms to return
    :param max_by_species: Limit to a maximum number of atoms by species
    :param max_force_error: In order to avoid counting in highly unlikely
        configurations, if the error exceeds this, do not add atom
    :return: Bool indicating if any atoms exceeded the error
        threshold, and a list of indices of atoms which did sorted by their
        error.
    """

    # Always returns true; use this when you want to test model performance
    # without updating the training set.
    if abs_force_tolerance == 0:
        return True, [-1]

    errors = np.abs(predicted_forces - label_forces)

    # Determine if any force component will trigger the threshold
    max_error_components = np.amax(errors, axis=1)
    if np.nanmax(max_error_components) < abs_force_tolerance:
        return True, [-1]

    target_atoms = []

    # Sort from greatest to smallest error
    force_arg_sorted = np.flip(np.argsort(max_error_components))

    present_species = {spec: 0 for spec in set(structure.species_labels)}

    # Only add atoms up to the bound
    for i in force_arg_sorted:

        # If max atoms added reached or force errors are now below threshold,
        # conclude
        if len(target_atoms) == max_atoms_added or (
            max_error_components[i] < abs_force_tolerance
            and max_error_components[i] != np.nan
        ):
            break

        cur_spec = structure.species_labels[i]

        # Only add up to species allowance, if it exists
        if (
            present_species[cur_spec] < max_by_species.get(cur_spec, inf)
            and max_error_components[i] < max_force_error
        ):
            target_atoms.append(i)
            present_species[cur_spec] += 1

    # Check in case that nothing was added e.g. due to species or error
    # limitations
    if len(target_atoms):
        return False, target_atoms
    else:
        return True, [-1]


def subset_of_frame_by_element(
    frame: "flare.Structure", predict_atoms_per_element: dict
) -> List[int]:
    """
    Given a structure and a dictionary formatted as {"Symbol":int,
    ..} describing a number of atoms per element, return a sorted list of
    indices corresponding to a random subset of atoms by species
    :param frame:
    :param predict_atoms_by_species:
    :return:
    """

    # Null case: No dictionary or empty dict passed in; just return all indices
    if not predict_atoms_per_element:
        return list(range(len(frame)))

    # Keep track of atoms which were considered (dictionary may only cover a
    #  subset of species of the whole frame)
    all_atoms = set(range(len(frame)))
    return_atoms = []
    considered_atoms = set([])

    species = frame.species_labels

    # Main loop: Obtain the number of relevant atoms for each element
    for elt, n in predict_atoms_per_element.items():

        matching_atoms = [i for i in all_atoms if species[i] == elt]
        considered_atoms.update(matching_atoms)

        if len(matching_atoms) == 0:
            continue
        # Choose the atoms to add
        to_add_atoms = np.random.choice(
            matching_atoms, replace=False, size=min(n, len(matching_atoms))
        )
        return_atoms += list(to_add_atoms)

    return_atoms += list(all_atoms - considered_atoms)

    return_atoms.sort()

    return return_atoms


def get_max_cutoff(cell: np.ndarray) -> float:
    """Compute the maximum cutoff compatible with a 3x3x3 supercell of a
        structure. Called in the Structure constructor when
        setting the max_cutoff attribute, which is used to create local
        environments with arbitrarily large cutoff radii.

    Args:
        cell (np.ndarray): Bravais lattice vectors of the structure stored as
            rows of a 3x3 Numpy array.

    Returns:
        float: Maximum cutoff compatible with a 3x3x3 supercell of the
            structure.
    """

    # Retrieve the lattice vectors.
    a_vec = cell[0]
    b_vec = cell[1]
    c_vec = cell[2]

    # Compute dot products and norms of lattice vectors.
    a_dot_b = np.dot(a_vec, b_vec)
    a_dot_c = np.dot(a_vec, c_vec)
    b_dot_c = np.dot(b_vec, c_vec)

    a_norm = np.linalg.norm(a_vec)
    b_norm = np.linalg.norm(b_vec)
    c_norm = np.linalg.norm(c_vec)

    # Compute the six independent altitudes of the cell faces.
    # The smallest is the maximum atomic environment cutoff that can be
    # used with sweep=1.
    max_candidates = np.zeros(6)
    max_candidates[0] = a_norm * np.sqrt(1 - (a_dot_b / (a_norm * b_norm)) ** 2)
    max_candidates[1] = b_norm * np.sqrt(1 - (a_dot_b / (a_norm * b_norm)) ** 2)
    max_candidates[2] = a_norm * np.sqrt(1 - (a_dot_c / (a_norm * c_norm)) ** 2)
    max_candidates[3] = c_norm * np.sqrt(1 - (a_dot_c / (a_norm * c_norm)) ** 2)
    max_candidates[4] = b_norm * np.sqrt(1 - (b_dot_c / (b_norm * c_norm)) ** 2)
    max_candidates[5] = c_norm * np.sqrt(1 - (b_dot_c / (b_norm * c_norm)) ** 2)

    return np.min(max_candidates)



"""
The :class:`Structure` object is a collection of atoms in a periodic box.
The mandatory inputs are the cell vectors of the box and the chemical species
and *Cartesian coordinates* of the atoms.
The atoms are automatically folded back into the primary cell, so the
input coordinates don't need to lie inside the box.
"""

from typing import List, Union, Any
from json import dumps, loads

import pickle as pickle
import numpy as np

from flare.utils.element_coder import element_to_Z, Z_to_element, NumpyEncoder
from flare.utils.learner import get_max_cutoff

try:
    # Used for to_pmg_structure method
    import pymatgen.core.structure as pmgstruc
    import pymatgen.io.vasp.inputs as pmgvaspio

    _pmg_present = True
except ImportError:
    _pmg_present = False


class Structure:
    """
    Contains information about a periodic structure of atoms, including the
    periodic cell boundaries, atomic species, and coordinates.

    *Note that input positions are assumed to be Cartesian.*

    :param cell: 3x3 array whose rows are the Bravais lattice vectors of the
        cell.
    :type cell: np.ndarray
    :param species: List of atomic species, which are represented either as
        integers or chemical symbols.
    :type species: List
    :param positions: Nx3 array of atomic coordinates.
    :type positions: np.ndarray
    :param mass_dict: Dictionary of atomic masses used in MD simulations.
    :type mass_dict: dict
    :param prev_positions: Nx3 array of previous atomic coordinates used in
        MD simulations.
    :type prev_positions: np.ndarray
    :param species_labels: List of chemical symbols. Used in the output file
        of on-the-fly runs.
    :type species_labels: List[str]
    :param stds: Uncertainty associated with forces
    :type stds: np.ndarray
    """

    def __init__(
        self,
        cell: "ndarray",
        species: Union[List[str], List[int]],
        positions: "ndarray",
        mass_dict: dict = None,
        prev_positions: "ndarray" = None,
        species_labels: List[str] = None,
        forces=None,
        stds=None,
        energy: float = None,
    ):

        # Define cell (each row is a Bravais lattice vector).
        self.cell = np.array(cell)

        # Compute the max cutoff compatible with a 3x3x3 supercell of the
        # structure.
        self.max_cutoff = get_max_cutoff(self.cell)

        # Set positions.
        self.positions = np.array(positions)

        # If species are strings, convert species to integers by atomic number
        if species_labels is None:
            self.species_labels = species
        else:
            self.species_labels = species_labels
        self.coded_species = np.array([element_to_Z(spec) for spec in species])
        self.nat = len(species)

        # Default: atoms have no velocity
        if prev_positions is None:
            self.prev_positions = np.copy(self.positions)
        else:
            assert len(positions) == len(
                prev_positions
            ), "Previous positions and positions are not same length"
            self.prev_positions = prev_positions

        # Set forces, energies, and stresses and their uncertainties.
        if forces is not None:
            self.forces = np.array(forces)
        else:
            self.forces = np.zeros((len(positions), 3))

        if stds is not None:
            self.stds = np.array(stds)
        else:
            self.stds = np.zeros((len(positions), 3))

        self.energy = energy

        self.local_energies = None
        self.local_energy_stds = None
        self.partial_stresses = None
        self.partial_stress_stds = None
        self.stress = None
        self.stress_stds = None

        # Potential energy attribute needed to mirror ASE atoms object.
        self.potential_energy = None

        self.mass_dict = mass_dict

        # Convert from elements to atomic numbers in mass dict
        if mass_dict is not None:
            keys = list(mass_dict.keys())
            for elt in keys:
                if isinstance(elt, str):
                    mass_dict[element_to_Z(elt)] = mass_dict[elt]
                    if elt.isnumeric():
                        mass_dict[int(elt)] = mass_dict[elt]

    @property
    def positions(self):
        return self._positions

    @property
    def wrapped_positions(self):
        return self._wrapped_positions

    @positions.setter
    def positions(self, position_array):
        self._positions = position_array
        self._wrapped_positions = self.wrap_positions()

    @property
    def cell(self):
        return self._cell

    @property
    def vec1(self):
        return self._vec1

    @property
    def vec2(self):
        return self._vec2

    @property
    def vec3(self):
        return self._vec3

    @property
    def cell_transpose(self):
        return self._cell_transpose

    @property
    def cell_transpose_inverse(self):
        return self._cell_transpose_inverse

    @property
    def cell_dot(self):
        return self._cell_dot

    @property
    def cell_dot_inverse(self):
        return self._cell_dot_inverse

    @cell.setter
    def cell(self, cell_array):
        """Set the cell and related properties."""
        self._cell = cell_array
        self._vec1 = cell_array[0, :]
        self._vec2 = cell_array[1, :]
        self._vec3 = cell_array[2, :]
        self._cell_transpose = cell_array.transpose()
        self._cell_transpose_inverse = np.linalg.inv(self._cell_transpose)
        self._cell_dot = self.get_cell_dot(cell_array)
        self._cell_dot_inverse = np.linalg.inv(self._cell_dot)

    @staticmethod
    def get_cell_dot(cell_array):
        """
        Compute 3x3 array of dot products of cell vectors used to
        fold atoms back to the unit cell.

        :return: 3x3 array of cell vector dot products.
        :rtype: np.ndarray
        """

        cell_dot = np.zeros((3, 3))

        for m in range(3):
            for n in range(3):
                cell_dot[m, n] = np.dot(cell_array[m], cell_array[n])

        return cell_dot

    @staticmethod
    def raw_to_relative(
        positions: "ndarray", cell_transpose: "ndarray", cell_dot_inverse: "ndarray"
    ) -> "ndarray":
        """Convert Cartesian coordinates to relative (fractional) coordinates,
        expressed in terms of the cell vectors set in self.cell.

        :param positions: Cartesian coordinates.
        :type positions: np.ndarray
        :param cell_transpose: Transpose of the cell array.
        :type cell_transpose: np.ndarray
        :param cell_dot_inverse: Inverse of the array of dot products of
            cell vectors.
        :type cell_dot_inverse: np.ndarray
        :return: Relative positions.
        :rtype: np.ndarray
        """

        relative_positions = np.matmul(
            np.matmul(positions, cell_transpose), cell_dot_inverse
        )

        return relative_positions

    @staticmethod
    def relative_to_raw(
        relative_positions: "ndarray",
        cell_transpose_inverse: "ndarray",
        cell_dot: "ndarray",
    ) -> "ndarray":
        """Convert fractional coordinates to raw (Cartesian) coordinates.

        :param relative_positions: fractional coordinates.
        :type relative_positions: np.ndarray
        :param cell_transpose_inverse: Transpose of the cell array.
        :type cell_transpose_inverse: np.ndarray
        :param cell_dot: Dot products of cell vectors
        :type cell_dot: np.ndarray
        :return: Cartesian positions.
        :rtype: np.ndarray
        """

        return np.matmul(
            np.matmul(relative_positions, cell_dot), cell_transpose_inverse
        )

    def wrap_positions(self) -> "ndarray":
        """
        Convenience function which folds atoms outside of the unit cell back
        into the unit cell. in_place flag controls if the wrapped positions
        are set in the class.

        :return: Cartesian coordinates of positions all in unit cell
        :rtype: np.ndarray
        """
        rel_pos = self.raw_to_relative(
            self.positions, self.cell_transpose, self.cell_dot_inverse
        )

        rel_wrap = rel_pos - np.floor(rel_pos)

        pos_wrap = self.relative_to_raw(
            rel_wrap, self.cell_transpose_inverse, self.cell_dot
        )

        return pos_wrap

    def indices_of_specie(self, specie: Union[int, str]) -> List[int]:
        """
        Return the indices of a given species within atoms of the structure.

        :param specie: Element to target, can be string or integer
        :return: The indices in the structure at which this element occurs
        :rtype: List[str]
        """
        return [i for i, spec in enumerate(self.coded_species) if spec == specie]

    # TODO make more descriptive
    def __str__(self) -> str:
        """
        Simple descriptive string of structure.

        :return: One-line descriptor of number of atoms and species present.
        :rtype: str
        """

        return "Structure with {} atoms of types {}".format(
            self.nat, set(self.species_labels)
        )

    def __len__(self) -> int:
        """
        Returns number of atoms in structure.

        :return: number of atoms in structure.
        :rtype: int
        """
        return self.nat

    def as_dict(self) -> dict:
        """
        Returns structure as a dictionary; useful for serialization purposes.

        :return: Dictionary version of current structure
        :rtype: dict
        """
        return dict(vars(self))

    def as_str(self) -> str:
        """
        Returns string dictionary serialization cast as string.

        :return: output of as_dict method cast as string
        :rtype: str
        """
        return dumps(self.as_dict(), cls=NumpyEncoder)

    @staticmethod
    def from_dict(dictionary: dict) -> "flare.struc.Structure":
        """
        Assembles a Structure object from a dictionary parameterizing one.

        :param dictionary: dict describing structure parameters.
        :return: FLARE structure assembled from dictionary
        """
        struc = Structure(
            cell=np.array(dictionary.get("_cell", dictionary.get("cell"))),
            positions=np.array(
                dictionary.get("_positions", dictionary.get("positions"))
            ),
            species=dictionary["coded_species"],
            forces=np.array(dictionary.get("forces")),
            mass_dict=dictionary.get("mass_dict"),
            species_labels=dictionary.get("species_labels"),
            energy=dictionary.get("energy", None),
        )

        struc.stds = np.array(dictionary.get("stds"))

        return struc

    @staticmethod
    def from_ase_atoms(atoms: "ase.Atoms", cell=None) -> "flare.struc.Structure":
        """
        From an ASE Atoms object, return a FLARE structure

        :param atoms: ASE Atoms object
        :type atoms: ASE Atoms object
        :return: A FLARE structure from an ASE atoms object
        """

        if cell is None:
            cell = np.array(atoms.cell)

        try:
            forces = atoms.get_forces()
        except:
            forces = None

        try:
            stds = atoms.get_uncertainties()
        except:
            stds = None

        try:
            energy = atoms.get_potential_energy()
        except:
            energy = None

        try:
            stress = atoms.get_stress()
        except:
            stress = None

        struc = Structure(
            cell=cell,
            positions=atoms.positions,
            species=atoms.get_chemical_symbols(),
            forces=forces,
            stds=stds,
            energy=energy,
        )
        struc.stress = stress
        return struc

    def to_ase_atoms(self) -> "ase.Atoms":
        from ase import Atoms
        from ase.calculators.singlepoint import SinglePointCalculator

        atoms = Atoms(
            self.species_labels, positions=self.positions, cell=self.cell, pbc=True
        )

        results = {}
        properties = ["forces", "energy", "stress"]
        for p in properties:
            results[p] = getattr(self, p)
        calculator = SinglePointCalculator(atoms, **results)
        atoms.set_calculator(calculator)

        return atoms

    def to_pmg_structure(self):
        """
        Returns FLARE structure as a pymatgen structure.

        :return: Pymatgen structure corresponding to current FLARE structure
        """

        if not _pmg_present:
            raise ModuleNotFoundError(
                "Pymatgen is not present. Please install Pymatgen and try again"
            )

        if self.forces is None:
            forces_temp = np.zeros((len(self.positions), 3))
            site_properties = {"force:": forces_temp, "std": self.stds}
        else:
            site_properties = {"force:": self.forces, "std": self.stds}

        return pmgstruc.Structure(
            lattice=self.cell,
            species=self.species_labels,
            coords=self.positions,
            coords_are_cartesian=True,
            site_properties=site_properties,
        )

    @staticmethod
    def from_pmg_structure(structure: "pymatgen Structure") -> "flare Structure":
        """
        Returns Pymatgen structure as FLARE structure.

        :param structure: Pymatgen Structure
        :type structure: Pymatgen Structure
        :return: FLARE Structure
        """

        cell = structure.lattice.matrix.copy()
        species = [str(spec) for spec in structure.species]
        positions = structure.cart_coords.copy()

        new_struc = Structure(cell=cell, species=species, positions=positions)

        site_props = structure.site_properties

        if "force" in site_props.keys():
            forces = site_props["force"]
            new_struc.forces = [np.array(force) for force in forces]

        if "std" in site_props.keys():
            stds = site_props["std"]
            new_struc.stds = [np.array(std) for std in stds]

        return new_struc

    def is_valid(self, tolerance: float = 0.5) -> bool:
        """
        Plugin to pymatgen's is_valid method to gauge if a structure
        has atoms packed too closely together, which is likely
        to cause unphysically large forces / energies or present convergence
        issues in DFT.
        :return:
        """
        pmg_structure = self.to_pmg_structure()
        return pmg_structure.is_valid(tol=tolerance)

    def to_xyz(
        self,
        extended_xyz: bool = True,
        print_stds: bool = False,
        print_forces: bool = False,
        print_max_stds: bool = False,
        print_energies: bool = False,
        predict_energy=None,
        dft_forces=None,
        dft_energy=None,
        timestep=-1,
        write_file: str = "",
        append: bool = False,
    ) -> str:
        """
        Convenience function which turns a structure into an extended .xyz
        file; useful for further input into visualization programs like VESTA
        or Ovito. Can be saved to an output file via write_file.

        :param print_stds: Print the stds associated with the structure.
        :param print_forces:
        :param extended_xyz:
        :param print_max_stds:
        :param write_file:
        :return:
        """
        species_list = [Z_to_element(x) for x in self.coded_species]
        xyz_str = ""
        xyz_str += f"{len(self.coded_species)} \n"

        # Add header line with info about lattice and properties if extended
        #  xyz option is called.
        if extended_xyz:
            cell = self.cell

            xyz_str += f'Lattice="{cell[0,0]} {cell[0,1]} {cell[0,2]}'
            xyz_str += f" {cell[1,0]} {cell[1,1]} {cell[1,2]}"
            xyz_str += f' {cell[2,0]} {cell[2,1]} {cell[2,2]}"'
            if timestep > 0:
                xyz_str += f" Timestep={timestep}"
            if predict_energy:
                xyz_str += f" PE={predict_energy}"
            if dft_energy is not None:
                xyz_str += f" DFT_PE={dft_energy}"
            xyz_str += f' Proprties="species:S:1:pos:R:3'

            if print_stds:
                xyz_str += ":stds:R:3"
                stds = self.stds
            if print_forces:
                xyz_str += ":forces:R:3"
                forces = self.forces
            if print_max_stds:
                xyz_str += ":max_std:R:1"
                stds = self.stds
            if print_energies:
                if self.local_energies is None:
                    print_energies = False
                else:
                    xyz_str += ":local_energy:R:1"
                    local_energies = self.local_energies
            if dft_forces is not None:
                xyz_str += ":dft_forces:R:3"
            xyz_str += "\n"
        else:
            xyz_str += "\n"

        for i, pos in enumerate(self.positions):
            # Write positions
            xyz_str += f"{species_list[i]} {pos[0]} {pos[1]} {pos[2]}"

            # If extended XYZ: Add in extra information
            if print_stds and extended_xyz:
                xyz_str += f" {stds[i,0]} {stds[i,1]} {stds[i,2]}"
            if print_forces and extended_xyz:
                xyz_str += f" {forces[i,0]} {forces[i,1]} {forces[i,2]}"
            if print_energies and extended_xyz:
                xyz_str += f" {local_energies[i]}"
            if print_max_stds and extended_xyz:
                xyz_str += f" {np.max(stds[i,:])} "
            if dft_forces is not None:
                xyz_str += f" {dft_forces[i, 0]} {dft_forces[i,1]} {dft_forces[i, 2]}"
            if i < (len(self.positions) - 1):
                xyz_str += "\n"

        # Write to file, optionally
        if write_file:
            if append:
                fmt = "a"
            else:
                fmt = "w"
            with open(write_file, fmt) as f:
                f.write(xyz_str)
                f.write("\n")

        return xyz_str

    @staticmethod
    def from_file(
        file_name: str, format: str = ""
    ) -> Union["flare.struc.Structure", List["flare.struc.Structure"]]:
        """
        Load a FLARE structure from a file or a series of FLARE structures
        :param file_name:
        :param format:
        :return:
        """

        # Ensure the file specified exists.
        with open(file_name, "r") as _:
            pass

        if "pickle" in file_name or (
            "pickle" in format.lower() or "binary" in format.lower()
        ):
            with open(file_name, "rb") as f:
                return pickle.load(f)

        if "xyz" in file_name or "xyz" in format.lower():
            raise NotImplementedError

        if "json" in format.lower() or ".json" in file_name:
            # Assumed format is one FLARE structure per line,
            # or one line with many FLARE structures
            with open(file_name, "r") as f:
                thelines = f.readlines()

                non_empty_lines = [loads(line) for line in thelines if len(line) > 2]

            structures = [
                Structure.from_dict(struc_dict) for struc_dict in non_empty_lines
            ]

            if len(structures) == 1:
                return structures[0]
            else:
                return structures

        is_poscar = (
            "POSCAR" in file_name or "CONTCAR" in file_name or "vasp" in format.lower()
        )
        if is_poscar and _pmg_present:
            pmg_structure = pmgvaspio.Poscar.from_file(file_name).structure
            return Structure.from_pmg_structure(pmg_structure)
        elif is_poscar and not _pmg_present:
            raise ImportError("Pymatgen not imported; functionality requires pymatgen.")


def get_unique_species(species: List[Any]) -> (List, List[int]):
    """
    Returns a list of the unique species passed in, and a list of
    integers indexing them.

    :param species: Species to index uniquely
    :return: List of the unique species, and integer indexes
    """
    unique_species = []
    coded_species = []
    for spec in species:
        if spec in unique_species:
            coded_species.append(unique_species.index(spec))
        else:
            coded_species.append(len(unique_species))
            unique_species.append(spec)
    coded_species = np.array(coded_species)

    return unique_species, coded_species
